from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

from numbskull import NumbSkull
from numbskull.inference import FACTORS
from numbskull.numbskulltypes import Weight, Variable, Factor, FactorToVar
import numpy as np
import random
import scipy.sparse as sparse
from fonduer.snorkel.learning import GenerativeModel, GenerativeModelWeights


def generate_model(n, dep_density, class_prior=False, lf_propensity=False, lf_prior=False, lf_class_propensity=False,
                   dep_similar=False, dep_reinforcing=False, dep_fixing=False, dep_exclusive=False, force_dep=False):
    weights = GenerativeModelWeights(n)
    for i in range(n):
        weights.lf_accuracy[i] = 1.1 - 0.2 * random.random()

    if class_prior:
        weights.class_prior = random.choice((-1.0, -2.0))

    if lf_propensity:
        for i in range(n):
            weights.lf_propensity[i] = random.choice((-1.0, -2.0))

    if lf_prior:
        for i in range(n):
            weights.lf_prior[i] = random.choice((1.0, -1.0))

    if lf_class_propensity:
        for i in range(n):
            weights.lf_class_propensity[i] = random.choice((1.0, -1.0))

    if dep_similar:
        for i in range(n):
            for j in range(i+1, n):
                if random.random() < dep_density:
                    weights.dep_similar[i, j] = 0.25

    if dep_fixing:
        for i in range(n):
            for j in range(i+1, n):
                if random.random() < dep_density:
                    if random.random() < 0.5:
                        weights.dep_fixing[i, j] = 0.25
                    else:
                        weights.dep_fixing[j, i] = 0.25

    if dep_reinforcing:
        for i in range(n):
            for j in range(i+1, n):
                if random.random() < dep_density:
                    if random.random() < 0.5:
                        weights.dep_reinforcing[i, j] = 0.25
                    else:
                        weights.dep_reinforcing[j, i] = 0.25

    if dep_exclusive:
        for i in range(n):
            for j in range(i+1, n):
                if random.random() < dep_density:
                    weights.dep_exclusive[i, j] = 0.25

    if force_dep and weights.dep_similar.getnnz() == 0 and weights.dep_fixing.getnnz() == 0 \
        and weights.dep_reinforcing.getnnz() == 0 and weights.dep_exclusive.getnnz() == 0:
        return generate_model(n, dep_density, class_prior=class_prior, lf_propensity=lf_propensity, lf_prior=lf_prior,
                              lf_class_propensity=lf_class_propensity, dep_similar=dep_similar, dep_fixing=dep_fixing,
                              dep_reinforcing=dep_reinforcing, dep_exclusive=dep_exclusive, force_dep=True)
    else:
        return weights


def generate_label_matrix(weights, m):
    # Compilation

    # Weights
    n_weights = 1 if weights.class_prior != 0.0 else 0

    n_weights += weights.n

    for optional_name in GenerativeModel.optional_names:
        for i in range(weights.n):
            if getattr(weights, optional_name)[i] != 0.0:
                n_weights += 1

    for dep_name in GenerativeModel.dep_names:
        for i in range(weights.n):
            for j in range(weights.n):
                if getattr(weights, dep_name)[i, j] != 0.0:
                    n_weights += 1

    weight = np.zeros(n_weights, Weight)
    for i in range(len(weight)):
        weight[i]['isFixed'] = True

    if weights.class_prior != 0.0:
        weight[0]['initialValue'] = np.float64(weights.class_prior)
        w_off = 1
    else:
        w_off = 0

    for i in range(weights.n):
        weight[w_off + i]['initialValue'] = np.float64(weights.lf_accuracy[i])
    w_off += weights.n

    for optional_name in GenerativeModel.optional_names:
        for i in range(weights.n):
            if getattr(weights, optional_name)[i] != 0.0:
                weight[w_off]['initialValue'] = np.float64(getattr(weights, optional_name)[i])
                w_off += 1

    for dep_name in GenerativeModel.dep_names:
        for i in range(weights.n):
            for j in range(weights.n):
                if getattr(weights, dep_name)[i, j] != 0.0:
                    weight[w_off]['initialValue'] = np.float64(getattr(weights, dep_name)[i, j])
                    w_off += 1

    # Variables
    variable = np.zeros(1 + weights.n, Variable)

    variable[0]['isEvidence'] = 0
    variable[0]['initialValue'] = 0
    variable[0]["dataType"] = 0
    variable[0]["cardinality"] = 2

    for i in range(weights.n):
        variable[1 + i]['isEvidence'] = 0
        variable[1 + i]['initialValue'] = 0
        variable[1 + i]["dataType"] = 0
        variable[1 + i]["cardinality"] = 3

    # Factors and FactorToVar
    n_edges = 1 if weights.class_prior != 0.0 else 0
    n_edges += 2 * weights.n
    for optional_name in GenerativeModel.optional_names:
        for i in range(weights.n):
            if getattr(weights, optional_name)[i] != 0.0:
                if optional_name == 'lf_prior' or optional_name == 'lf_propensity':
                    n_edges += 1
                elif optional_name == 'lf_class_propensity':
                    n_edges += 2
                else:
                    raise ValueError()
    for dep_name in GenerativeModel.dep_names:
        for i in range(weights.n):
            for j in range(weights.n):
                if getattr(weights, dep_name)[i, j] != 0.0:
                    if dep_name == 'dep_similar' or dep_name == 'dep_exclusive':
                        n_edges += 2
                    elif dep_name == 'dep_fixing' or dep_name == 'dep_reinforcing':
                        n_edges += 3
                    else:
                        raise ValueError()

    factor = np.zeros(n_weights, Factor)
    ftv = np.zeros(n_edges, FactorToVar)

    if weights.class_prior != 0.0:
        factor[0]["factorFunction"] = FACTORS["DP_GEN_CLASS_PRIOR"]
        factor[0]["weightId"] = 0
        factor[0]["featureValue"] = 1
        factor[0]["arity"] = 1
        factor[0]["ftv_offset"] = 0

        ftv[0]["vid"] = 0

        f_off = 1
        ftv_off = 1
    else:
        f_off = 0
        ftv_off = 0

    for i in range(weights.n):
        factor[f_off + i]["factorFunction"] = FACTORS["DP_GEN_LF_ACCURACY"]
        factor[f_off + i]["weightId"] = f_off + i
        factor[f_off + i]["featureValue"] = 1
        factor[f_off + i]["arity"] = 2
        factor[f_off + i]["ftv_offset"] = ftv_off + 2 * i

        ftv[ftv_off + 2 * i]["vid"] = 0
        ftv[ftv_off + 2 * i + 1]["vid"] = 1 + i
    f_off += weights.n
    ftv_off += 2 * weights.n

    for i in range(weights.n):
        if weights.lf_prior[i] != 0.0:
            factor[f_off]["factorFunction"] = FACTORS["DP_GEN_LF_PRIOR"]
            factor[f_off]["weightId"] = f_off
            factor[f_off]["featureValue"] = 1
            factor[f_off]["arity"] = 1
            factor[f_off]["ftv_offset"] = ftv_off

            ftv[ftv_off]["vid"] = 1 + i
            f_off += 1
            ftv_off += 1

    for i in range(weights.n):
        if weights.lf_propensity[i] != 0.0:
            factor[f_off]["factorFunction"] = FACTORS["DP_GEN_LF_PROPENSITY"]
            factor[f_off]["weightId"] = f_off
            factor[f_off]["featureValue"] = 1
            factor[f_off]["arity"] = 1
            factor[f_off]["ftv_offset"] = ftv_off

            ftv[ftv_off]["vid"] = 1 + i
            f_off += 1
            ftv_off += 1

    for i in range(weights.n):
        if weights.lf_class_propensity[i] != 0.0:
            factor[f_off]["factorFunction"] = FACTORS["DP_GEN_LF_CLASS_PROPENSITY"]
            factor[f_off]["weightId"] = f_off
            factor[f_off]["featureValue"] = 1
            factor[f_off]["arity"] = 2
            factor[f_off]["ftv_offset"] = ftv_off

            ftv[ftv_off]["vid"] = 0
            ftv[ftv_off + 1]["vid"] = 1 + i

            f_off += 1
            ftv_off += 2

    for dep_name in GenerativeModel.dep_names:
        for i in range(weights.n):
            for j in range(weights.n):
                if getattr(weights, dep_name)[i, j] != 0.0:
                    if dep_name == 'dep_similar' or dep_name == 'dep_exclusive':
                        factor[f_off]["factorFunction"] = FACTORS["DP_GEN_DEP_SIMILAR"] if dep_name == 'dep_similar' else FACTORS["DP_GEN_DEP_EXCLUSIVE"]
                        factor[f_off]["weightId"] = f_off
                        factor[f_off]["featureValue"] = 1
                        factor[f_off]["arity"] = 2
                        factor[f_off]["ftv_offset"] = ftv_off

                        ftv[ftv_off]["vid"] = 1 + i
                        ftv[ftv_off + 1]["vid"] = 1 + j

                        f_off += 1
                        ftv_off += 2
                    elif dep_name == 'dep_fixing' or dep_name == 'dep_reinforcing':
                        factor[f_off]["factorFunction"] = FACTORS["DP_GEN_DEP_FIXING"] if dep_name == 'dep_fixing' else FACTORS["DP_GEN_DEP_REINFORCING"]

                        factor[f_off]["weightId"] = f_off
                        factor[f_off]["featureValue"] = 1
                        factor[f_off]["arity"] = 3
                        factor[f_off]["ftv_offset"] = ftv_off

                        ftv[ftv_off]["vid"] = 0
                        ftv[ftv_off + 1]["vid"] = 1 + i
                        ftv[ftv_off + 2]["vid"] = 1 + j

                        f_off += 1
                        ftv_off += 3
                    else:
                        raise ValueError()

    # Domain mask
    domain_mask = np.zeros(1 + weights.n, np.bool)

    # Instantiates factor graph
    ns = NumbSkull(n_inference_epoch=100, quiet=True)
    ns.loadFactorGraph(weight, variable, factor, ftv, domain_mask, n_edges)
    fg = ns.getFactorGraph()

    y = np.ndarray((m,), np.int64)
    L = sparse.lil_matrix((m, weights.n), dtype=np.int64)
    for i in range(m):
        fg.burnIn(10, False)
        y[i] = 1 if fg.var_value[0, 0] == 0 else -1
        for j in range(weights.n):
            if fg.var_value[0, 1 + j] != 2:
                L[i, j] = 1 if fg.var_value[0, 1 + j] == 0 else -1

    return y, L.tocsr()



from fonduer.features.content_features import *
from fonduer.features.core_features import *
from fonduer.features.structural_features import *
from fonduer.features.table_features import *
from fonduer.features.visual_features import *


def get_all_feats(candidates):
    for id, f, v in get_core_feats(candidates):
        yield id, f, v
    for id, f, v in get_content_feats(candidates):
        yield id, f, v
    for id, f, v in get_structural_feats(candidates):
        yield id, f, v
    for id, f, v in get_table_feats(candidates):
        yield id, f, v
    for id, f, v in get_visual_feats(candidates):
        yield id, f, v



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

import numpy as np

from .rnn_base import RNNBase
from .utils import candidate_to_tokens, SymbolTable


def tag(seq, labels):
    assert(len(seq) == len(labels))
    seq_new, t = [], False
    for x, y in zip(seq, labels):
        if y and (not t):
            seq_new.append(self.OPEN)
            seq_new.append(x)
            t = True
        elif (not y) and t:
            seq_new.append(self.CLOSE)
            seq_new.append(x)
            t = False
        else:
            seq_new.append(x)
    return seq_new


class TagRNN(RNNBase):
    """TagRNN for sequence tagging"""
    OPEN, CLOSE = '~~[[~~', '~~]]~~'

    def _preprocess_data(self, candidates, extend=False):
        """Convert candidate sentences to tagged symbol sequences
            @candidates: candidates to process
            @extend: extend symbol table for tokens (train), or lookup (test)?
        """
        if not hasattr(self, 'word_dict'):
            self.word_dict = SymbolTable()
        data, ends = [], []
        for candidate in candidates:
            # Read sentence data
            tokens = candidate_to_tokens(candidate)
            # Get label sequence
            labels = np.zeros(len(tokens), dtype=int)
            labels[c[0].get_word_start() : c[0].get_word_end()+1] = 1
            # Tag sequence
            s = tag(tokens, labels)
            # Either extend word table or retrieve from it
            f = self.word_dict.get if extend else self.word_dict.lookup
            data.append(np.array(list(map(f, s))))
            ends.append(c[0].get_word_end())
        return data, ends



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

import numpy as np
import tensorflow as tf

from fonduer.snorkel.learning.disc_learning import TFNoiseAwareModel
from fonduer.snorkel.learning.utils import LabelBalancer, reshape_marginals
from scipy.sparse import csr_matrix, issparse
from six.moves.cPickle import dump, load
from time import time

SD = 0.1


class LogisticRegression(TFNoiseAwareModel):
    representation = False

    def _build_model(self, d=None, **kwargs):
        """
        Build model, setting logits and marginals ops.

        :param d: Number of features
        """
        s1, s2 = self.seed, (self.seed + 1 if self.seed is not None else None)

        # Define inputs
        self.X = tf.placeholder(tf.float32, (None, d))
        self.Y = tf.placeholder(tf.float32, (None, self.cardinality)) \
            if self.cardinality > 2 else tf.placeholder(tf.float32, (None,))

        # Define parameters and logits
        k = self.cardinality if self.cardinality > 2 else 1
        self.w = tf.Variable(tf.random_normal((d, k), stddev=SD, seed=s1))
        self.b = tf.Variable(tf.random_normal((k,), stddev=SD, seed=s2))

        if self.deterministic:
            # TODO: Implement for categorical as well...
            if self.cardinality > 2:
                raise NotImplementedError(
                    "Deterministic mode not implemented for categoricals.")

            # Make deterministic
            # See: https://www.twosigma.com/insights/a-workaround-for-non-determinism-in-tensorflow
            f_w = tf.matmul(self.X, self.w)
            f_w_temp = tf.concat([f_w, tf.ones_like(f_w)], axis=1)
            b_temp = tf.stack([tf.ones_like(self.b), self.b], axis=0)
            self.logits = tf.matmul(f_w_temp, b_temp)
        else:
            self.logits = tf.nn.bias_add(tf.matmul(self.X, self.w), self.b)

        if self.cardinality == 2:
            self.logits = tf.squeeze(self.logits)

        # Define marginals op
        marginals_fn = tf.nn.softmax if self.cardinality > 2 else tf.nn.sigmoid
        self.marginals_op = marginals_fn(self.logits)

    def _build_training_ops(self, l1_penalty=0.0, l2_penalty=0.0, **kwargs):
        """
        Build training ops, setting loss and train ops

        :param l1_penalty: L1 reg. coefficient
        :param l2_penalty: L2 reg. coefficient
        """
        super(LogisticRegression, self)._build_training_ops()

        # Add L1 and L2 penalties
        if l1_penalty > 0:
            self.loss += l1_penalty * tf.reduce_sum(tf.abs(self.w))
        if l2_penalty > 0:
            self.loss += l2_penalty * tf.nn.l2_loss(self.w)

    def _construct_feed_dict(self, X_b, Y_b, lr=0.01, **kwargs):
        return {self.X: X_b, self.Y: Y_b, self.lr: lr}

    def _check_input(self, X):
        if issparse(X):
            raise Exception("Sparse input matrix. Use SparseLogisticRegression")
        return X

    def _marginals_batch(self, X_test):
        X_test = self._check_input(X_test)
        return self.session.run(self.marginals_op, {self.X: X_test})

    def get_weights(self):
        """Get model weights and bias"""
        w, b = self.session.run([self.w, self.b])
        return np.ravel(w), np.ravel(b)


class SparseLogisticRegression(LogisticRegression):
    representation = False

    def _build_model(self, d=None, **kwargs):
        # Define sparse input placeholders + sparse tensors
        self.indices = tf.placeholder(tf.int64)
        self.shape   = tf.placeholder(tf.int64, (2,))
        self.ids     = tf.placeholder(tf.int64)
        self.weights = tf.placeholder(tf.float32)
        sparse_ids   = tf.SparseTensor(self.indices, self.ids, self.shape)
        sparse_vals  = tf.SparseTensor(self.indices, self.weights, self.shape)
        self.Y = tf.placeholder(tf.float32, (None, self.cardinality)) \
            if self.cardinality > 2 else tf.placeholder(tf.float32, (None,))

        # Define parameters and logits
        s1, s2 = self.seed, (self.seed + 1 if self.seed is not None else None)
        k = self.cardinality if self.cardinality > 2 else 1
        self.w = tf.Variable(tf.random_normal((d, k), stddev=SD, seed=s1))
        self.b = tf.Variable(tf.random_normal((k,), stddev=SD, seed=s2))

        if self.deterministic:
            # TODO: Implement for categorical as well...
            if self.cardinality > 2:
                raise NotImplementedError(
                    "Deterministic mode not implemented for categoricals.")

            # Try to make deterministic...
            f_w = tf.nn.embedding_lookup_sparse(params=self.w,
                sp_ids=sparse_ids, sp_weights=sparse_vals, combiner=None)
            f_w_temp = tf.concat([f_w, tf.ones_like(f_w)], axis=1)
            b_temp = tf.stack([tf.ones_like(self.b), self.b], axis=0)
            self.logits = tf.matmul(f_w_temp, b_temp)
        else:
            z = tf.nn.embedding_lookup_sparse(params=self.w, sp_ids=sparse_ids,
                sp_weights=sparse_vals, combiner='sum')
            self.logits = tf.nn.bias_add(z, self.b)

        if self.cardinality == 2:
            self.logits = tf.squeeze(self.logits)

        # Define marginals op
        marginals_fn = tf.nn.softmax if self.cardinality > 2 else tf.nn.sigmoid
        self.marginals_op = marginals_fn(self.logits)

    def _check_input(self, X):
        if not issparse(X):
            msg = "Dense input matrix. Cast to sparse or use LogisticRegression"
            raise Exception(msg)
        return X.tocsr()

    def _batch_sparse_data(self, X):
        """
        Convert sparse batch matrix to sparse inputs for embedding lookup
        Notes: https://github.com/tensorflow/tensorflow/issues/342
        """
        if not issparse(X):
            raise Exception("Matrix X must be scipy.sparse type")
        X_lil = X.tolil()
        indices, ids, weights = [], [], []
        max_len = 0
        for i, (row, data) in enumerate(zip(X_lil.rows, X_lil.data)):
            # Dummy weight for all-zero row
            if len(row) == 0:
                indices.append((i, 0))
                ids.append(0)
                weights.append(0.0)
                continue
            # Update indices by position
            max_len = max(max_len, len(row))
            indices.extend((i, t) for t in range(len(row)))
            ids.extend(row)
            weights.extend(data)
        shape = (len(X_lil.rows), max_len)
        return indices, shape, ids, weights

    def _construct_feed_dict(self, X_b, Y_b, lr=0.01, **kwargs):
        indices, shape, ids, weights = self._batch_sparse_data(X_b)
        return {
            self.indices: indices,
            self.shape:   shape,
            self.ids:     ids,
            self.weights: weights,
            self.Y:       Y_b,
            self.lr:      lr
        }

    def _marginals_batch(self, X_test):
        X_test = self._check_input(X_test)
        if X_test.shape[0] == 0:
            return np.array([])
        indices, shape, ids, weights = self._batch_sparse_data(X_test)
        return self.session.run(self.marginals_op, {
            self.indices: indices,
            self.shape:   shape,
            self.ids:     ids,
            self.weights: weights
        })



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

from fonduer.snorkel.models import Candidate, Context, Sentence
from fonduer.snorkel.udf import UDF, UDFRunner


class CorpusParser(UDFRunner):
    def __init__(self, parser=None, fn=None):
        self.parser = parser
        super(CorpusParser, self).__init__(
            CorpusParserUDF, parser=self.parser, fn=fn)

    def clear(self, session, **kwargs):
        session.query(Context).delete()
        # We cannot cascade up from child contexts to parent Candidates,
        # so we delete all Candidates too
        session.query(Candidate).delete()


class CorpusParserUDF(UDF):
    def __init__(self, parser, fn, **kwargs):
        super(CorpusParserUDF, self).__init__(**kwargs)
        self.parser = parser
        self.req_handler = parser.connect()
        self.fn = fn

    def apply(self, x, **kwargs):
        """Given a Document object and its raw text, parse into Sentences"""
        doc, text = x
        for parts in self.req_handler.parse(doc, text):
            parts = self.fn(parts) if self.fn is not None else parts
            yield Sentence(**parts)



import json
import os
import re
import six
import lxml.etree as et
from fonduer.snorkel.utils import corenlp_cleaner

APP_HOME = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))

# Load IPython display functionality libs if possible i.e. if in IPython
try:
    from IPython.core.display import display_html, HTML, display_javascript, Javascript
except:
    pass


class XMLTree:
    """
    A generic tree representation which takes XML as input
    Includes subroutines for conversion to JSON & for visualization based on js form
    """

    def __init__(self, xml_root, words=None):
        """Calls subroutines to generate JSON form of XML input"""
        self.root = xml_root
        self.words = words

        # create a unique id for e.g. canvas id in notebook
        self.id = str(abs(hash(self.to_str())))

    def _to_json(self, root):
        js = {'attrib': dict(root.attrib), 'children': []}
        for i, c in enumerate(root):
            js['children'].append(self._to_json(c))
        return js

    def to_json(self):
        return self._to_json(self.root)

    def to_str(self):
        return et.tostring(self.root)

    def render_tree(self, highlight=[]):
        """
        Renders d3 visualization of the d3 tree, for IPython notebook display
        Depends on html/js files in vis/ directory, which is assumed to be in same dir...
        """
        # HTML
        WORD = '<span class="word-' + self.id + '-%s">%s</span>'
        words = ' '.join(WORD % (i, w) for i, w in enumerate(
            corenlp_cleaner(self.words))) if self.words else ''
        html = open('%s/vis/tree-chart.html' % APP_HOME).read() % (self.id,
                                                                   self.id,
                                                                   words)
        display_html(HTML(data=html))

        # JS
        JS_LIBS = ["https://d3js.org/d3.v3.min.js"]
        js = open('%s/vis/tree-chart.js' % APP_HOME).read() % (
            self.id, json.dumps(self.to_json()), str(highlight))
        display_javascript(Javascript(data=js, lib=JS_LIBS))


def corenlp_to_xmltree(s, prune_root=True):
    """
    Transforms an object with CoreNLP dep_path and dep_parent attributes into an XMLTree
    Will include elements of any array having the same dimensiion as dep_* as node attributes
    Also adds special word_idx attribute corresponding to original sequence order in sentence
    """
    # Convert input object to dictionary
    if isinstance(s, dict) == False:
        try:
            s = s.__dict__ if hasattr(s, '__dict__') else dict(s)
        except:
            raise ValueError("Cannot convert input object to dict")

    # Use the dep_parents array as a guide: ensure it is present and a list of ints
    if not ('dep_parents' in s and isinstance(s['dep_parents'], list)):
        raise ValueError(
            "Input CoreNLP object must have a 'dep_parents' attribute which is a list"
        )
    try:
        dep_parents = list(map(int, s['dep_parents']))
    except:
        raise ValueError("'dep_parents' attribute must be a list of ints")

    # Also ensure that we are using CoreNLP-native indexing (root=0, 1-base word indexes)!
    b = min(dep_parents)
    if b != 0:
        dep_parents = list(map(lambda j: j - b, dep_parents))

    # Parse recursively
    root = corenlp_to_xmltree_sub(s, dep_parents, 0)

    # Often the return tree will have several roots, where one is the actual root
    # And the rest are just singletons not included in the dep tree parse...
    # We optionally remove these singletons and then collapse the root if only one child left
    if prune_root:
        for c in root:
            if len(c) == 0:
                root.remove(c)
        if len(root) == 1:
            root = root.findall("./*")[0]
    return XMLTree(root, words=s['words'])


def scrub(s):
    return ''.join(c for c in s if ord(c) < 128)


def corenlp_to_xmltree_sub(s, dep_parents, rid=0):
    i = rid - 1
    attrib = {}
    N = len(list(dep_parents))

    # Add all attributes that have the same shape as dep_parents
    if i >= 0:
        for k, v in list(
                filter(lambda t: isinstance(t[1], list) and len(t[1]) == N,
                       six.iteritems(s))):
            if v[i] is not None:
                attrib[singular(k)] = scrub(v[i]).encode(
                    'ascii', 'ignore') if hasattr(v[i], 'encode') else str(
                        v[i])

        # Add word_idx if not present
        if 'word_idx' not in attrib:
            attrib['word_idx'] = str(i)

    # Build tree recursively
    root = et.Element('node', attrib=attrib)
    for i, d in enumerate(dep_parents):
        if d == rid:
            root.append(corenlp_to_xmltree_sub(s, dep_parents, i + 1))
    return root


def singular(s):
    """Get singular form of word s (crudely)"""
    return re.sub(r'e?s$', '', s, flags=re.I)



"""
Subpackage for learning the structures of models.
"""
from __future__ import absolute_import

from .gen_learning import DependencySelector
from .synthetic import generate_model, generate_label_matrix
from .utils import *



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

import codecs
import glob
import os
import re
import lxml.etree as et

from bs4 import BeautifulSoup

from ..models import Document


class DocPreprocessor(object):
    """
    Processes a file or directory of files into a set of Document objects.

    :param encoding: file encoding to use, default='utf-8'
    :param path: filesystem path to file or directory to parse
    :param max_docs: the maximum number of Documents to produce,
        default=float('inf')

    """

    def __init__(self, path, encoding="utf-8", max_docs=float('inf')):
        self.path = path
        self.encoding = encoding
        self.max_docs = max_docs

    def generate(self):
        """
        Parses a file or directory of files into a set of Document objects.

        """
        doc_count = 0
        for fp in self._get_files(self.path):
            file_name = os.path.basename(fp)
            if self._can_read(file_name):
                for doc, text in self.parse_file(fp, file_name):
                    yield doc, text
                    doc_count += 1
                    if doc_count >= self.max_docs:
                        return

    def __iter__(self):
        return self.generate()

    def get_stable_id(self, doc_id):
        return "%s::document:0:0" % doc_id

    def parse_file(self, fp, file_name):
        raise NotImplementedError()

    def _can_read(self, fpath):
        return not fpath.startswith('.')

    def _get_files(self, path):
        if os.path.isfile(path):
            fpaths = [path]
        elif os.path.isdir(path):
            fpaths = [os.path.join(path, f) for f in os.listdir(path)]
        else:
            fpaths = glob.glob(path)
        if len(fpaths) > 0:
            return fpaths
        else:
            raise IOError("File or directory not found: %s" % (path,))


class TSVDocPreprocessor(DocPreprocessor):
    """Simple parsing of TSV file with one (doc_name <tab> doc_text) per line"""

    def parse_file(self, fp, file_name):
        with codecs.open(fp, encoding=self.encoding) as tsv:
            for line in tsv:
                (doc_name, doc_text) = line.split('\t')
                stable_id = self.get_stable_id(doc_name)
                doc = Document(
                    name=doc_name, stable_id=stable_id,
                    meta={'file_name': file_name}
                )
                yield doc, doc_text


class TextDocPreprocessor(DocPreprocessor):
    """Simple parsing of raw text files, assuming one document per file"""

    def parse_file(self, fp, file_name):
        with codecs.open(fp, encoding=self.encoding) as f:
            name = os.path.basename(fp).rsplit('.', 1)[0]
            stable_id = self.get_stable_id(name)
            doc = Document(
                name=name, stable_id=stable_id, meta={'file_name': file_name}
            )
            yield doc, f.read()


class CSVPathsPreprocessor(DocPreprocessor):
    """This `DocumentPreprocessor` treats inputs file as index of paths to
     actual documents; each line in the input file contains a path to a document.

     **Defaults and Customization:**

     * The input file is treated as a simple text file having one path per file.
       However, if the input is a CSV file, a pair of ``column`` and ``delim``
       parameters may be used to retrieve the desired value as reference path.

     * The referenced documents are treated as text document and hence parsed
       using ``TextDocPreprocessor``. However, if the referenced files are
       complex, an advanced parser may be used by specifying ``parser_factory``
       parameter to constructor.
     """

    def __init__(self, path, parser_factory=TextDocPreprocessor, column=None,
                 delim=',', *args, **kwargs):
        """
        :param path: input file having paths
        :param parser_factory: The parser class to be used to parse the
            referenced files. default = TextDocPreprocessor
        :param column: index of the column which references path.
                 default=None, which implies that each line has only one column
        :param delim: delimiter to be used to separate columns when file has
                      more than one column. It is active only when
                      ``column is not None``. default=','
        """
        super(CSVPathsPreprocessor, self).__init__(path, *args, **kwargs)
        self.column = column
        self.delim = delim
        self.parser = parser_factory(path)

    def _get_files(self, path):
        with codecs.open(path, encoding=self.encoding) as lines:
            for doc_path in lines:
                if self.column is not None:
                    # if column is set, retrieve specific column from CSV record
                    doc_path = doc_path.split(self.delim)[self.column]
                yield doc_path.strip()

    def parse_file(self, fp, file_name):
        return self.parser.parse_file(fp, file_name)


class HTMLDocPreprocessor(DocPreprocessor):
    """Simple parsing of raw HTML files, assuming one document per file"""

    def parse_file(self, fp, file_name):
        with open(fp, 'rb') as f:
            html = BeautifulSoup(f, 'lxml')
            txt = list(filter(self._cleaner, html.findAll(text=True)))
            txt = ' '.join(self._strip_special(s) for s in txt if s != '\n')
            name = os.path.basename(fp).rsplit('.', 1)[0]
            stable_id = self.get_stable_id(name)
            doc = Document(
                name=name, stable_id=stable_id, meta={'file_name': file_name}
            )
            yield doc, txt

    def _can_read(self, fpath):
        return fpath.endswith('.html')

    def _cleaner(self, s):
        if s.parent.name in ['style', 'script', '[document]', 'head', 'title']:
            return False
        elif re.match('<!--.*-->', str(s)):
            return False
        return True

    def _strip_special(self, s):
        return (''.join(c for c in s if ord(c) < 128)).encode('ascii', 'ignore')


class XMLMultiDocPreprocessor(DocPreprocessor):
    """
    Parse an XML file _which contains multiple documents_ into a set of Document
    objects.

    Use XPath queries to specify a _document_ object, and then for each
    document, a set of _text_ sections and an _id_.

    **Note: Include the full document XML etree in the attribs dict with
    keep_xml_tree=True**
    """

    def __init__(self, path, doc='.//document', text='./text/text()',
        id='./id/text()', keep_xml_tree=False, *args, **kwargs):
        super(XMLMultiDocPreprocessor, self).__init__(path, *args, **kwargs)
        self.doc = doc
        self.text = text
        self.id = id
        self.keep_xml_tree = keep_xml_tree

    def parse_file(self, f, file_name):
        for i, doc in enumerate(et.parse(f).xpath(self.doc)):
            doc_id = str(doc.xpath(self.id)[0])
            text = '\n'.join(
                [t for t in doc.xpath(self.text) if t is not None]
            )
            meta = {'file_name': str(file_name)}
            if self.keep_xml_tree:
                meta['root'] = et.tostring(doc)
            stable_id = self.get_stable_id(doc_id)
            yield Document(name=doc_id, stable_id=stable_id, meta=meta), text

    def _can_read(self, fpath):
        return fpath.endswith('.xml')



"""
Subpackage for all built-in Snorkel data models.

After creating additional models that extend snorkel.models.meta.SnorkelBase, they must be added to the database schema.
For example, the correct way to define a new type of Context is:

.. code-block:: python

    from fonduer.snorkel.models.context import Context
    from sqlalchemy import Column, String, ForeignKey

    class NewType(Context):
        # Declares name for storage table
        __tablename__ = 'newtype'
        # Connects NewType records to generic Context records
        id = Column(String, ForeignKey('context.id', ondelete='CASCADE'), primary_key=True)

        # Polymorphism information for SQLAlchemy
        __mapper_args__ = {
            'polymorphic_identity': 'newtype',
        }

        # Rest of class definition here

    # Adds the corresponding table to the underlying database's schema
    from fonduer.snorkel.models.meta import SnorkelBase, snorkel_engine
    SnorkelBase.metadata.create_all(snorkel_engine)
"""
from __future__ import absolute_import

from .meta import SnorkelBase, SnorkelSession, snorkel_engine, snorkel_postgres
from .context import Context, Document, Sentence, TemporarySpan, Span
from .context import construct_stable_id, split_stable_id
from .candidate import Candidate, candidate_subclass, Marginal
from .annotation import (
    Feature, FeatureKey, Label, LabelKey, GoldLabel, GoldLabelKey, StableLabel,
    Prediction, PredictionKey
)

# This call must be performed after all classes that extend SnorkelBase are
# declared to ensure the storage schema is initialized
SnorkelBase.metadata.create_all(snorkel_engine)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

import tensorflow as tf
import numpy as np
from time import time
import os
from six.moves.cPickle import dump, load

from .classifier import Classifier
from .utils import reshape_marginals, LabelBalancer

class TFNoiseAwareModel(Classifier):
    """
    Generic NoiseAwareModel class for TensorFlow models.
    Note that the actual network is built when train is called (to allow for
    model architectures which depend on the training data, e.g. vocab size).
    
    :param n_threads: Parallelism to use; single-threaded if None
    :param seed: Top level seed which is passed into both numpy operations
        via a RandomState maintained by the class, and into TF as a graph-level
        seed.
    :param deterministic [EXPERIMENTAL / in development!] If True, attempts to
            make the model deterministic on GPU by replacing all reduce_ and 
            other non-deterministic operations; has no effect (other than 
            potential slight slowdown) for CPU (at least for single-threaded?).
    """
    def __init__(self, n_threads=None, seed=123, deterministic=False, **kwargs):
        self.n_threads = n_threads
        self.seed = seed
        self.rand_state = np.random.RandomState()
        self.deterministic = deterministic
        super(TFNoiseAwareModel, self).__init__(**kwargs)

    def _build_model(self, **model_kwargs):
        """
        Builds the TensorFlow Operations for the model. Must set the following:
            - self.logits: The un-normalized potentials for the variables
                ("logits" in keeping with TensorFlow terminology)
            - self.Y: The training marginals to fit to
            - self.marginals_op: Normalized predicted marginals for the vars

        Additional ops must be set depending on whether the default
        self._construct_feed_dict method below is used, or a custom one.

        Note that _build_model is called in the train method, allowing for the 
        network to be constructed dynamically depending on the training set 
        (e.g., the size of the vocabulary for a text LSTM model)

        Note also that model_kwargs are saved to disk by the self.save method,
        as they are needed to rebuild / reload the model. *All hyperparameters
        needed to rebuild the model must be passed in here for model reloading
        to work!*
        """
        raise NotImplementedError()

    def _build_training_ops(self, **training_kwargs):
        """
        Builds the TensorFlow Operations for the training procedure. Must set 
        the following:
            - self.loss: Loss function
            - self.optimizer: Training operation
        """
        # Define loss and marginals ops
        if self.cardinality > 2:
            loss_fn = tf.nn.softmax_cross_entropy_with_logits
        else:
            loss_fn = tf.nn.sigmoid_cross_entropy_with_logits

        # If deterministic=True, avoid use of non-deterministic reduce_ ops
        if self.deterministic:
            l = tf.reshape(loss_fn(logits=self.logits, labels=self.Y), [1, -1])
            self.loss = tf.squeeze(tf.matmul(l, tf.ones_like(l), 
                transpose_b=True)) / tf.cast(tf.shape(l)[1], tf.float32)
        else:
            self.loss = tf.reduce_mean(loss_fn(logits=self.logits, 
                labels=self.Y))
        
        # Build training op
        self.lr = tf.placeholder(tf.float32)
        self.optimizer = tf.train.AdamOptimizer(self.lr).minimize(self.loss)

    def _construct_feed_dict(self, X_b, Y_b, lr=0.01, **kwargs):
        """
        Given a batch of data and labels, and other necessary hyperparams,
        construct a python dictionary to use in the training step as feed_dict.

        This method can be overwritten when non-standard arguments are needed.

        NOTE: Using a feed_dict is obviously not the fastest way to pass in data
            if running a large-scale dataset on a GPU, see Issue #679.
        """
        # Note: The below arguments need to be constructed by self._build!
        return {self.X: X_b, self.Y: Y_b, self.lr: lr}

    def _build_new_graph_session(self, **model_kwargs):
        """Creates new graph, builds network, and sets new session."""
        self.model_kwargs = model_kwargs

        # Create new computation graph
        self.graph = tf.Graph()

        with self.graph.as_default():

            # Set graph-level random seed
            tf.set_random_seed(self.seed)

            # Build network here in the graph
            self._build_model(**model_kwargs)

        # Create new session
        self.session = tf.Session(
            config=tf.ConfigProto(
                intra_op_parallelism_threads=self.n_threads,
                inter_op_parallelism_threads=self.n_threads
            ),
            graph=self.graph
        ) if self.n_threads is not None else tf.Session(graph=self.graph)

    def _check_input(self, X):
        """Checks correctness of input; optional to implement."""
        pass

    def train(self, X_train, Y_train, n_epochs=25, lr=0.01, batch_size=256, 
        rebalance=False, X_dev=None, Y_dev=None, print_freq=5, dev_ckpt=True,
        dev_ckpt_delay=0.75, save_dir='checkpoints', **kwargs):
        """
        Generic training procedure for TF model

        :param X_train: The training Candidates. If self.representation is True, then
            this is a list of Candidate objects; else is a csr_AnnotationMatrix
            with rows corresponding to training candidates and columns 
            corresponding to features.
        :param Y_train: Array of marginal probabilities for each Candidate
        :param n_epochs: Number of training epochs
        :param lr: Learning rate
        :param batch_size: Batch size for SGD
        :param rebalance: Bool or fraction of positive examples for training
                    - if True, defaults to standard 0.5 class balance
                    - if False, no class balancing
        :param X_dev: Candidates for evaluation, same format as X_train
        :param Y_dev: Labels for evaluation, same format as Y_train
        :param print_freq: number of epochs at which to print status, and if present,
            evaluate the dev set (X_dev, Y_dev).
        :param dev_ckpt: If True, save a checkpoint whenever highest score
            on (X_dev, Y_dev) reached. Note: currently only evaluates at
            every @print_freq epochs.
        :param dev_ckpt_delay: Start dev checkpointing after this portion
            of n_epochs.
        :param save_dir: Save dir path for checkpointing.
        :param kwargs: All hyperparameters that change how the graph is built 
            must be passed through here to be saved and reloaded to save /
            reload model. *NOTE: If a parameter needed to build the 
            network and/or is needed at test time is not included here, the
            model will not be able to be reloaded!*
        """
        self._check_input(X_train)
        verbose = print_freq > 0

        # Set random seed for all numpy operations
        self.rand_state.seed(self.seed)

        # If the data passed in is a feature matrix (representation=False),
        # set the dimensionality here; else assume this is done by sub-class
        if not self.representation:
            kwargs['d'] = X_train.shape[1]

        # Check that the cardinality of the training marginals and model agree
        cardinality = Y_train.shape[1] if len(Y_train.shape) > 1 else 2
        if cardinality != self.cardinality:
            raise ValueError("Training marginals cardinality ({0}) does not"
                "match model cardinality ({1}).".format(Y_train.shape[1], 
                    self.cardinality))
        # Make sure marginals are in correct default format
        Y_train = reshape_marginals(Y_train)
        # Make sure marginals are in [0,1] (v.s e.g. [-1, 1])
        if self.cardinality > 2 and not np.all(Y_train.sum(axis=1) - 1 < 1e-10):
            raise ValueError("Y_train must be row-stochastic (rows sum to 1).")
        if not np.all(Y_train >= 0):
            raise ValueError("Y_train must have values in [0,1].")

        # Remove unlabeled examples (i.e. P(X=k) == 1 / cardinality for all k)
        # and optionally rebalance training set
        # Note: rebalancing only for binary setting currently
        if self.cardinality == 2:
            # This removes unlabeled examples and optionally rebalances
            train_idxs = LabelBalancer(Y_train).get_train_idxs(rebalance,
                rand_state=self.rand_state)
        else:
            # In categorical setting, just remove unlabeled
            diffs = Y_train.max(axis=1) - Y_train.min(axis=1)
            train_idxs = np.where(diffs > 1e-6)[0]
        X_train = [X_train[j] for j in train_idxs] if self.representation \
            else X_train[train_idxs, :]
        Y_train = Y_train[train_idxs]

        # Create new graph, build network, and start session
        self._build_new_graph_session(**kwargs)

        # Build training ops
        # Note that training_kwargs and model_kwargs are mixed together; ideally
        # would be separated but no negative effect
        with self.graph.as_default():
            self._build_training_ops(**kwargs)
        
        # Initialize variables
        with self.graph.as_default():
            self.session.run(tf.global_variables_initializer())

        # Run mini-batch SGD
        n = len(X_train) if self.representation else X_train.shape[0]
        batch_size = min(batch_size, n)
        if verbose:
            st = time()
            print("[{0}] Training model".format(self.name))
            print("[{0}] n_train={1}  #epochs={2}  batch size={3}".format(
                self.name, n, n_epochs, batch_size
            ))
        dev_score_opt = 0.0
        for t in range(n_epochs):
            epoch_losses = []
            for i in range(0, n, batch_size):
                feed_dict = self._construct_feed_dict(
                    X_train[i:min(n, i+batch_size)],
                    Y_train[i:min(n, i+batch_size)],
                    lr=lr,
                    **kwargs
                )
                # Run training step and evaluate loss function    
                epoch_loss, _ = self.session.run(
                    [self.loss, self.optimizer], feed_dict=feed_dict)
                epoch_losses.append(epoch_loss)

            # Reshuffle training data
            train_idxs = self.rand_state.permutation(list(range(n)))
            X_train = [X_train[j] for j in train_idxs] if self.representation \
                else X_train[train_idxs, :]
            Y_train = Y_train[train_idxs]
            
            # Print training stats and optionally checkpoint model
            if verbose and (t % print_freq == 0 or t in [0, (n_epochs-1)]):
                msg = "[{0}] Epoch {1} ({2:.2f}s)\tAverage loss={3:.6f}".format(
                    self.name, t, time() - st, np.mean(epoch_losses))
                if X_dev is not None:
                    scores = self.score(X_dev, Y_dev, batch_size=batch_size)
                    score = scores if self.cardinality > 2 else scores[-1]
                    score_label = "Acc." if self.cardinality > 2 else "F1"
                    msg += '\tDev {0}={1:.2f}'.format(score_label, 100. * score)
                print(msg)
                    
                # If best score on dev set so far and dev checkpointing is
                # active, save checkpoint
                if X_dev is not None and dev_ckpt and \
                    t > dev_ckpt_delay * n_epochs and score > dev_score_opt:
                    dev_score_opt = score
                    self.save(save_dir=save_dir, global_step=t)
        
        # Conclude training
        if verbose:
            print("[{0}] Training done ({1:.2f}s)".format(self.name, time()-st))

        # If checkpointing on, load last checkpoint (i.e. best on dev set)
        if dev_ckpt and X_dev is not None and verbose and dev_score_opt > 0:
            self.load(save_dir=save_dir)

    def marginals(self, X, batch_size=None):
        """
        Compute the marginals for the given candidates X.
        Split into batches to avoid OOM errors, then call _marginals_batch;
        defaults to no batching.
        """
        if batch_size is None:
            return self._marginals_batch(X)
        else:
            N = len(X) if self.representation else X.shape[0]
            n_batches = int(np.floor(N / batch_size))

            # Iterate over batches
            batch_marginals = []
            for b in range(0, N, batch_size):
                batch = self._marginals_batch(X[b:min(N, b+batch_size)])
                
                # Note: if a single marginal in *binary* classification is
                # returned, it will have shape () rather than (1,)- catch here
                if len(batch.shape) == 0:
                    batch = batch.reshape(1)
                    
                batch_marginals.append(batch)
            return np.concatenate(batch_marginals)

    def save(self, model_name=None, save_dir='checkpoints', verbose=True,
        global_step=0):
        """Save current model."""
        model_name = model_name or self.name

        # Note: Model checkpoints need to be saved in separate directories!
        model_dir = os.path.join(save_dir, model_name)
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)

        # Create Saver
        with self.graph.as_default():
            saver = tf.train.Saver(tf.global_variables())

        # Save model kwargs needed to rebuild model
        with open(os.path.join(model_dir, "model_kwargs.pkl"), 'wb') as f:
            dump(self.model_kwargs, f)

        # Save graph and report if verbose
        saver.save(
            self.session,
            os.path.join(model_dir, model_name),
            global_step=global_step
        )
        if verbose:
            print("[{0}] Model saved as <{1}>".format(self.name, model_name))

    def load(self, model_name=None, save_dir='checkpoints', verbose=True):
        """Load model from file and rebuild in new graph / session."""
        model_name = model_name or self.name
        model_dir = os.path.join(save_dir, model_name)

        # Load model kwargs needed to rebuild model
        with open(os.path.join(model_dir, "model_kwargs.pkl"), 'rb') as f:
            model_kwargs = load(f)
        
        # Create new graph, build network, and start session
        self._build_new_graph_session(**model_kwargs)

        # Initialize variables
        with self.graph.as_default():
            self.session.run(tf.global_variables_initializer())

        # Load saved checkpoint to populate trained parameters
        with self.graph.as_default():
            saver = tf.train.Saver(tf.global_variables())
            ckpt = tf.train.get_checkpoint_state(model_dir)
        if ckpt and ckpt.model_checkpoint_path:
            saver.restore(self.session, ckpt.model_checkpoint_path)
            if verbose:
                print("[{0}] Loaded model <{1}>".format(self.name, model_name))
        else:
            raise Exception("[{0}] No model found at <{1}>".format(
                self.name, model_name))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

from sqlalchemy import (
    Column, String, Integer, Float, Boolean, ForeignKey, UniqueConstraint,
    MetaData
)
from sqlalchemy.orm import relationship, backref
from functools import partial

from .meta import SnorkelBase
from ..models import snorkel_engine
from ..utils import camel_to_under


class Candidate(SnorkelBase):
    """
    An abstract candidate relation.

    New relation types should be defined by calling candidate_subclass(),
    **not** subclassing this class directly.
    """
    __tablename__ = 'candidate'
    id          = Column(Integer, primary_key=True)
    type        = Column(String, nullable=False)
    split       = Column(Integer, nullable=False, default=0, index=True)

    __mapper_args__ = {
        'polymorphic_identity': 'candidate',
        'polymorphic_on': type
    }

    # __table_args__ = {"extend_existing" : True}

    def get_contexts(self):
        """Get a tuple of the consituent contexts making up this candidate"""
        return tuple(getattr(self, name) for name in self.__argnames__)

    def get_parent(self):
        # Fails if both contexts don't have same parent
        p = [c.get_parent() for c in self.get_contexts()]
        if p.count(p[0]) == len(p):
            return p[0]
        else:
            raise Exception("Contexts do not all have same parent")

    def get_cids(self):
        """Get a tuple of the canonical IDs (CIDs) of the contexts making up
        this candidate"""
        return tuple(getattr(self, name + "_cid") for name in self.__argnames__)

    def __len__(self):
        return len(self.__argnames__)

    def __getitem__(self, key):
        return self.get_contexts()[key]

    def __repr__(self):
        return "%s(%s)" % (
            self.__class__.__name__, ", ".join(map(str, self.get_contexts()))
            )

    def __gt__(self, other_cand):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other_cand.__repr__()

# This global dictionary contains all classes that have been declared in this Python environment, so
# that candidate_subclass() can return a class if it already exists and is identical in specification
# to the requested class
candidate_subclasses = {}

def candidate_subclass(class_name, args, table_name=None, cardinality=None,
    values=None):
    """
    Creates and returns a Candidate subclass with provided argument names,
    which are Context type. Creates the table in DB if does not exist yet.

    Import using:

    .. code-block:: python

        from fonduer.snorkel.models import candidate_subclass

    :param class_name: The name of the class, should be "camel case" e.g.
        NewCandidate
    :param args: A list of names of consituent arguments, which refer to the
        Contexts--representing mentions--that comprise the candidate
    :param table_name: The name of the corresponding table in DB; if not
        provided, is converted from camel case by default, e.g. new_candidate
    :param cardinality: The cardinality of the variable corresponding to the
        Candidate. By default is 2 i.e. is a binary value, e.g. is or is not
        a true mention.
    """
    if table_name is None:
        table_name = camel_to_under(class_name)

    # If cardinality and values are None, default to binary classification
    if cardinality is None and values is None:
        values = [True, False]
        cardinality = 2

    # Else use values if present, and validate proper input
    elif values is not None:
        if cardinality is not None and len(values) != cardinality:
            raise ValueError("Number of values must match cardinality.")
        if None in values:
            raise ValueError("`None` is a protected value.")
        # Note that bools are instances of ints in Python...
        if any([isinstance(v, int) and not isinstance(v, bool) for v in values]):
            raise ValueError("Default usage of values is consecutive integers. Leave values unset if attempting to define values as integers.")
        cardinality = len(values)

    # If cardinality is specified but not values, fill in with ints
    elif cardinality is not None:
        values = list(range(cardinality))

    class_spec = (args, table_name, cardinality, values)
    if class_name in candidate_subclasses:
        if class_spec == candidate_subclasses[class_name][1]:
            return candidate_subclasses[class_name][0]
        else:
            raise ValueError('Candidate subclass ' + class_name + ' already exists in memory with incompatible ' +
                             'specification: ' + str(candidate_subclasses[class_name][1]))
    else:
        # Set the class attributes == the columns in the database
        class_attribs = {

            # Declares name for storage table
            '__tablename__' : table_name,

            # Connects candidate_subclass records to generic Candidate records
            'id' : Column(
                Integer,
                ForeignKey('candidate.id', ondelete='CASCADE'),
                primary_key=True
            ),

            # Store values & cardinality information in the class only
            'values' : values,
            'cardinality' : cardinality,

            # Polymorphism information for SQLAlchemy
            '__mapper_args__' : {'polymorphic_identity': table_name},

            # Helper method to get argument names
            '__argnames__' : args,
        }

        # Create named arguments, i.e. the entity mentions comprising the relation
        # mention
        # For each entity mention: id, cid ("canonical id"), and pointer to Context
        unique_args = []
        for arg in args:

            # Primary arguments are constituent Contexts, and their ids
            class_attribs[arg + '_id'] = Column(
                Integer, ForeignKey('context.id', ondelete='CASCADE'))
            class_attribs[arg] = relationship(
                'Context',
                backref=backref(
                    table_name + '_' + arg + 's',
                    cascade_backrefs=False,
                    cascade='all, delete-orphan'
                ),
                cascade_backrefs=False,
                foreign_keys=class_attribs[arg + '_id']
            )
            unique_args.append(class_attribs[arg + '_id'])

            # Canonical ids, to be set post-entity normalization stage
            class_attribs[arg + '_cid'] = Column(String)

        # Add unique constraints to the arguments
        class_attribs['__table_args__'] = (
            UniqueConstraint(*unique_args),
        )

        # Create class
        C = type(class_name, (Candidate,), class_attribs)

        # Create table in DB
        if not snorkel_engine.dialect.has_table(snorkel_engine, table_name):
            C.__table__.create(bind=snorkel_engine)

        candidate_subclasses[class_name] = C, class_spec

        return C


class Marginal(SnorkelBase):
    """
    A marginal probability corresponding to a (Candidate, value) pair.

    Represents:

        P(candidate = value) = probability

    @training: If True, this is a training marginal; otherwise is end prediction
    """
    __tablename__ = 'marginal'
    id           = Column(Integer, primary_key=True)
    candidate_id = Column(Integer,
                        ForeignKey('candidate.id', ondelete='CASCADE'))
    training     = Column(Boolean, default=True)
    value        = Column(Integer, nullable=False, default=1)
    probability  = Column(Float, nullable=False, default=0.0)

    __table_args__ = (
        UniqueConstraint(candidate_id, training, value),
    )

    def __repr__(self):
        label = "Training" if self.training else "Predicted"
        return "<%s Marginal: P(%s == %s) = %s>" % \
            (label, self.candidate_id, self.value, self.probability)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

import logging
import os
import re

# Travis will not import the PorterStemmer
if 'CI' not in os.environ:
    try:
        from nltk.stem.porter import PorterStemmer
    except ImportError:
        logger = logging.getLogger(__name__)
        logger.warning("nltk not installed- some default functionality may be absent.")


class Matcher(object):
    """
    Applies a function f : c -> {True,False} to a generator of candidates,
    returning only candidates _c_ s.t. _f(c) == True_,
    where f can be compositionally defined.
    """
    def __init__(self, *children, **opts):
        self.children           = children
        self.opts               = opts
        self.longest_match_only = self.opts.get('longest_match_only', True)
        self.init()
        self._check_opts()

    def init(self):
        pass

    def _check_opts(self):
        """
        Checks for unsupported opts, throws error if found
        NOTE: Must be called _after_ init()
        """
        for opt in self.opts.keys():
            if not opt in self.__dict__:
                raise Exception("Unsupported option: %s" % opt)

    def _f(self, c):
        """The internal (non-composed) version of filter function f"""
        return True

    def f(self, c):
        """
        The recursively composed version of filter function f
        By default, returns logical **conjunction** of operator and single child operator
        """
        if len(self.children) == 0:
            return self._f(c)
        elif len(self.children) == 1:
            return self._f(c) and self.children[0].f(c)
        else:
            raise Exception("%s does not support more than one child Matcher" % self.__name__)

    def _is_subspan(self, c, span):
        """Tests if candidate c is subspan of span, where span is defined specific to candidate type"""
        return False

    def _get_span(self, c):
        """Gets a tuple that identifies a span for the specific candidate class that c belongs to"""
        return c

    def apply(self, candidates):
        """
        Apply the Matcher to a **generator** of candidates
        Optionally only takes the longest match (NOTE: assumes this is the *first* match)
        """
        seen_spans = set()
        for c in candidates:
            if self.f(c) and (not self.longest_match_only or not any([self._is_subspan(c, s) for s in seen_spans])):
                if self.longest_match_only:
                    seen_spans.add(self._get_span(c))
                yield c


WORDS = 'words'

class NgramMatcher(Matcher):
    """Matcher base class for Ngram objects"""
    def _is_subspan(self, c, span):
        """Tests if candidate c is subspan of span, where span is defined specific to candidate type"""
        return c.sentence.id == span[0] and c.char_start >= span[1] and c.char_end <= span[2]

    def _get_span(self, c):
        """Gets a tuple that identifies a span for the specific candidate class that c belongs to"""
        return (c.sentence.id, c.char_start, c.char_end)


class DictionaryMatch(NgramMatcher):
    """Selects candidate Ngrams that match against a given list d"""
    def init(self):
        self.ignore_case = self.opts.get('ignore_case', True)
        self.attrib      = self.opts.get('attrib', WORDS)
        self.reverse     = self.opts.get('reverse', False)
        try:
            self.d = frozenset(w.lower() if self.ignore_case else w for w in self.opts['d'])
        except KeyError:
            raise Exception("Please supply a dictionary (list of phrases) d as d=d.")

        # Optionally use a stemmer, preprocess the dictionary
        # Note that user can provide *an object having a stem() method*
        self.stemmer = self.opts.get('stemmer', None)
        if self.stemmer is not None:
            if self.stemmer == 'porter':
                self.stemmer = PorterStemmer()
            self.d = frozenset(self._stem(w) for w in list(self.d))

    def _stem(self, w):
        """Apply stemmer, handling encoding errors"""
        try:
            return self.stemmer.stem(w)
        except UnicodeDecodeError:
            return w

    def _f(self, c):
        p = c.get_attrib_span(self.attrib)
        p = p.lower() if self.ignore_case else p
        p = self._stem(p) if self.stemmer is not None else p
        return (not self.reverse) if p in self.d else self.reverse

class LambdaFunctionMatcher(NgramMatcher):
    """Selects candidate Ngrams that return True when fed to a function f."""
    def init(self):
        self.ignore_case = self.opts.get('ignore_case', True)
        self.attrib      = self.opts.get('attrib', WORDS)
        try:
            self.func = self.opts['func']
        except KeyError:
            raise Exception("Please supply a function f as func=f.")

    def _f(self, c):
        """The internal (non-composed) version of filter function f"""
        return self.func(c)


class Union(NgramMatcher):
    """Takes the union of candidate sets returned by child operators"""
    def f(self, c):
       for child in self.children:
           if child.f(c) > 0:
               return True
       return False


class Intersect(Matcher):
    """Takes the intersection of candidate sets returned by child operators"""
    def f(self, c):
        for child in self.children:
            if not child.f(c):
                return False
        return True


class Inverse(Matcher):
    """Returns the opposite result of its child operator"""
    # TODO: confirm that this only has one child
    def f(self, c):
        for child in self.children:
            return not child.f(c)


class Concat(NgramMatcher):
    """
    Selects candidates which are the concatenation of adjacent matches from child operators
    NOTE: Currently slices on **word index** and considers concatenation along these divisions only
    """
    def init(self):
        self.permutations   = self.opts.get('permutations', False)
        self.left_required  = self.opts.get('left_required', True)
        self.right_required = self.opts.get('right_required', True)
        self.ignore_sep     = self.opts.get('ignore_sep', True)
        self.sep            = self.opts.get('sep', " ")

    def f(self, c):
        if len(self.children) != 2:
            raise ValueError("Concat takes two child Matcher objects as arguments.")
        if not self.left_required and self.children[1].f(c):
            return True
        if not self.right_required and self.children[0].f(c):
            return True

        # Iterate over candidate splits **at the word boundaries**
        for wsplit in range(c.get_word_start()+1, c.get_word_end()+1):
            csplit = c.word_to_char_index(wsplit) - c.char_start  # NOTE the switch to **candidate-relative** char index

            # Optionally check for specific separator
            if self.ignore_sep or c.get_span()[csplit-1] == self.sep:
                c1 = c[:csplit-len(self.sep)]
                c2 = c[csplit:]
                if self.children[0].f(c1) and self.children[1].f(c2):
                    return True
                if self.permutations and self.children[1].f(c1) and self.children[0].f(c2):
                    return True
        return False


class SlotFillMatch(NgramMatcher):
    """Matches a slot fill pattern of matchers _at the character level_"""
    def init(self):
        self.attrib = self.opts.get('attrib', WORDS)
        try:
            self.pattern = self.opts['pattern']
        except KeyError:
            raise Exception("Please supply a slot-fill pattern p as pattern=p.")

        # Parse slot fill pattern
        split        = re.split(r'\{(\d+)\}', self.pattern)
        self._ops    = list(map(int, split[1::2]))
        self._splits = split[::2]

        # NOTE: Must have non-null splits!!
        if any([len(s) == 0 for s in self._splits[1:-1]]):
            raise ValueError("SlotFillMatch must have non-empty split patterns to function currently.")

        # Check for correct number of child matchers / slots
        if len(self.children) != len(set(self._ops)):
            raise ValueError("Number of provided matchers (%s) != number of slots (%s)." \
                    % (len(self.children), len(set(self._ops))))

    def f(self, c):

        # First, filter candidates by matching splits pattern
        m = re.match(r'(.+)'.join(self._splits) + r'$', c.get_attrib_span(self.attrib))
        if m is None:
            return False

        # Then, recursively apply matchers
        for i,op in enumerate(self._ops):
            if self.children[op].f(c[m.start(i+1):m.end(i+1)]) == 0:
                return False
        return True


class RegexMatch(NgramMatcher):
    """Base regex class- does not specify specific semantics of *what* is being matched yet"""
    def init(self):
        try:
            self.rgx = self.opts['rgx']
        except KeyError:
            raise Exception("Please supply a regular expression string r as rgx=r.")
        self.ignore_case = self.opts.get('ignore_case', True)
        self.attrib      = self.opts.get('attrib', WORDS)
        self.sep         = self.opts.get('sep', " ")

        # Compile regex matcher
        # NOTE: Enforce full span matching by ensuring that regex ends with $.
        # Group self.rgx first so that $ applies to all components of an 'OR'
        # expression. (e.g., we want r'(a|b)$' instead of r'a|b$')
        self.rgx = self.rgx if self.rgx.endswith('$') else ('(' + self.rgx + ')$')
        self.r = re.compile(self.rgx, flags=(re.I if self.ignore_case else 0) | re.UNICODE)

    def _f(self, c):
        raise NotImplementedError()


class RegexMatchSpan(RegexMatch):
    """Matches regex pattern on **full concatenated span**"""
    def _f(self, c):
        return True if self.r.match(c.get_attrib_span(self.attrib, sep=self.sep)) is not None else False


class RegexMatchEach(RegexMatch):
    """Matches regex pattern on **each token**"""
    def _f(self, c):
        tokens = c.get_attrib_tokens(self.attrib)
        return True if tokens and all([self.r.match(t) is not None for t in tokens]) else False


class PersonMatcher(RegexMatchEach):
    """
    Matches Spans that are the names of people, as identified by CoreNLP.

    A convenience class for setting up a RegexMatchEach to match spans
    for which each token was tagged as a person.
    """
    def __init__(self, *children, **kwargs):
        kwargs['attrib'] = 'ner_tags'
        kwargs['rgx'] = 'PERSON'
        super(PersonMatcher, self).__init__(*children, **kwargs)


class LocationMatcher(RegexMatchEach):
    """
    Matches Spans that are the names of locations, as identified by CoreNLP.

    A convenience class for setting up a RegexMatchEach to match spans
    for which each token was tagged as a location.
    """
    def __init__(self, *children, **kwargs):
        kwargs['attrib'] = 'ner_tags'
        kwargs['rgx'] = 'LOCATION|LOC'
        super(LocationMatcher, self).__init__(*children, **kwargs)


class OrganizationMatcher(RegexMatchEach):
    """
    Matches Spans that are the names of organizations, as identified by CoreNLP.

    A convenience class for setting up a RegexMatchEach to match spans
    for which each token was tagged as an organization.
    """
    def __init__(self, *children, **kwargs):
        kwargs['attrib'] = 'ner_tags'
        kwargs['rgx'] = 'ORGANIZATION|ORG'
        super(OrganizationMatcher, self).__init__(*children, **kwargs)


class DateMatcher(RegexMatchEach):
    """
    Matches Spans that are dates, as identified by CoreNLP.

    A convenience class for setting up a RegexMatchEach to match spans
    for which each token was tagged as a date.
    """
    def __init__(self, *children, **kwargs):
        kwargs['attrib'] = 'ner_tags'
        kwargs['rgx'] = 'DATE'
        super(DateMatcher, self).__init__(*children, **kwargs)


class NumberMatcher(RegexMatchEach):
    """
    Matches Spans that are numbers, as identified by CoreNLP.

    A convenience class for setting up a RegexMatchEach to match spans
    for which each token was tagged as a number.
    """
    def __init__(self, *children, **kwargs):
        kwargs['attrib'] = 'ner_tags'
        kwargs['rgx'] = 'NUMBER|QUANTITY'
        super(NumberMatcher, self).__init__(*children, **kwargs)


class MiscMatcher(RegexMatchEach):
    """
    Matches Spans that are miscellaneous named entities, as identified by CoreNLP.

    A convenience class for setting up a RegexMatchEach to match spans
    for which each token was tagged as miscellaneous.
    """
    def __init__(self, *children, **kwargs):
        kwargs['attrib'] = 'ner_tags'
        kwargs['rgx'] = 'MISC'
        super(MiscMatcher, self).__init__(*children, **kwargs)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

import re
import sys
import numpy as np
import scipy.sparse as sparse
import subprocess


class ProgressBar(object):
    def __init__(self, N, length=40):
        self.N      = max(1, N)
        self.nf     = float(self.N)
        self.length = length
        self.update_interval = self.nf/100
        self.current_tick = 0
        self.bar(0)

    def bar(self, i):
        """Assumes i ranges through [0, N-1]"""
        new_tick = i/self.update_interval
        if int(new_tick) != int(self.current_tick):
            b = int(np.ceil((i / self.nf) * self.length))
            sys.stdout.write("\r[%s%s] %d%%" % ("="*b, " "*(self.length-b), int(100*(i / self.nf))))
            sys.stdout.flush()
        self.current_tick = new_tick

    def close(self):
        b = self.length
        sys.stdout.write("\r[%s%s] %d%%\n" % ("="*b, " "*(self.length-b), 100))
        sys.stdout.flush()


def get_ORM_instance(ORM_class, session, instance):
    """
    Given an ORM class and *either an instance of this class, or the name attribute of an instance
    of this class*, return the instance
    """
    if isinstance(instance, str) or isinstance(instance, unicode):
        return session.query(ORM_class).filter(ORM_class.name == instance).one_or_none()
    else:
        return instance


def camel_to_under(name):
    """
    Converts camel-case string to lowercase string separated by underscores.

    Written by epost
    (http://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case).

    :param name: String to be converted
    :return: new String with camel-case converted to lowercase, underscored
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

def sparse_nonzero(X):
    """Sparse matrix with value 1 for i,jth entry !=0"""
    X_nonzero = X.copy()
    if not sparse.issparse(X):
        X_nonzero[X_nonzero != 0] = 1
        return X_nonzero
    if sparse.isspmatrix_csr(X) or sparse.isspmatrix_csc(X):
        X_nonzero.data[X_nonzero.data != 0] = 1
    elif sparse.isspmatrix_lil(X):
        X_nonzero.data = [np.ones(len(L)) for L in X_nonzero.data]
    else:
        raise ValueError("Only supports CSR/CSC and LIL matrices")
    return X_nonzero

def sparse_abs(X):
    """Element-wise absolute value of sparse matrix- avoids casting to dense matrix!"""
    X_abs = X.copy()
    if not sparse.issparse(X):
        return abs(X_abs)
    if sparse.isspmatrix_csr(X) or sparse.isspmatrix_csc(X):
        X_abs.data = np.abs(X_abs.data)
    elif sparse.isspmatrix_lil(X):
        X_abs.data = np.array([np.abs(L) for L in X_abs.data])
    else:
        raise ValueError("Only supports CSR/CSC and LIL matrices")
    return X_abs


def matrix_coverage(L):
    """
    Given an N x M matrix where L_{i,j} is the label given by the jth LF to the ith candidate:
    Return the **fraction of candidates that each LF labels.**
    """
    return np.ravel(sparse_nonzero(L).sum(axis=0) / float(L.shape[0]))


def matrix_overlaps(L):
    """
    Given an N x M matrix where L_{i,j} is the label given by the jth LF to the ith candidate:
    Return the **fraction of candidates that each LF _overlaps with other LFs on_.**
    """
    L_nonzero = sparse_nonzero(L)
    return np.ravel(np.where(L_nonzero.sum(axis=1) > 1, 1, 0).T * L_nonzero / float(L.shape[0]))

def matrix_conflicts(L):
    """
    Given an N x M matrix where L_{i,j} is the label given by the jth LF to the ith candidate:
    Return the **fraction of candidates that each LF _conflicts with other LFs on_.**
    """
    B = L.copy()
    if not sparse.issparse(B):
        for row in range(B.shape[0]):
            if np.unique(np.array(B[row][np.nonzero(B[row])])).size == 1:
                B[row] = 0
        return matrix_coverage(sparse_nonzero(B))
    if not (sparse.isspmatrix_csc(B) or sparse.isspmatrix_lil(B) or sparse.isspmatrix_csr(B)):
        raise ValueError("Only supports CSR/CSC and LIL matrices")
    if sparse.isspmatrix_csc(B) or sparse.isspmatrix_lil(B):
        B = B.tocsr()
    for row in range(B.shape[0]):
        if np.unique(B.getrow(row).data).size == 1:
            B.data[B.indptr[row]:B.indptr[row+1]] = 0
    return matrix_coverage(sparse_nonzero(B))



def matrix_tp(L, labels):
    return np.ravel([
        np.sum(np.ravel((L[:, j] == 1).todense()) * (labels == 1)) for j in range(L.shape[1])
    ])


def matrix_fp(L, labels):
    return np.ravel([
        np.sum(np.ravel((L[:, j] == 1).todense()) * (labels == -1)) for j in range(L.shape[1])
    ])


def matrix_tn(L, labels):
    return np.ravel([
        np.sum(np.ravel((L[:, j] == -1).todense()) * (labels == -1)) for j in range(L.shape[1])
    ])


def matrix_fn(L, labels):
    return np.ravel([
        np.sum(np.ravel((L[:, j] == -1).todense()) * (labels == 1)) for j in range(L.shape[1])
    ])


def get_as_dict(x):
    """Return an object as a dictionary of its attributes"""
    if isinstance(x, dict):
        return x
    else:
        try:
            return x._asdict()
        except AttributeError:
            return x.__dict__


def sort_X_on_Y(X, Y):
    return [x for (y,x) in sorted(zip(Y,X), key=lambda t : t[0])]


def corenlp_cleaner(words):
  d = {'-RRB-': ')', '-LRB-': '(', '-RCB-': '}', '-LCB-': '{',
       '-RSB-': ']', '-LSB-': '['}
  return [d[w] if w in d else w for w in words]


def split_html_attrs(attrs):
    """
    Given an iterable object of (attr, values) pairs, returns a list of separated
    "attr=value" strings
    """
    html_attrs = []
    for a in attrs:
        attr = a[0]
        values = [v.split(';') for v in a[1]] if isinstance(a[1], list) else [a[1].split(';')]
        for i in range(len(values)):
            while isinstance(values[i], list):
                values[i] = values[i][0]
        html_attrs += ["=".join([attr, val]) for val in values]
    return html_attrs


def tokens_to_ngrams(tokens, n_min=1, n_max=3, delim=' ', lower=False):
    f = (lambda x: x.lower()) if lower else (lambda x: x)
    N = len(tokens)
    for root in range(N):
        for n in range(max(n_min - 1, 0), min(n_max, N - root)):
            yield f(delim.join(tokens[root:root + n + 1]))


def get_keys_by_candidate(candidate, annotation_matrix):
    (r, c, v) = sparse.find(annotation_matrix[annotation_matrix.get_row_index(candidate), :])
    return [annotation_matrix.get_key(idx).name for idx in c]


def remove_files(filename):
    try:
        subprocess.check_output('rm -f %s' % filename, shell=True)
    except OSError as e:
        print(e)
        pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

from .meta import SnorkelBase, snorkel_postgres
import pickle
from sqlalchemy import Column, String, Integer, Text, ForeignKey, UniqueConstraint
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import relationship, backref
from sqlalchemy.sql import select, text
from sqlalchemy.types import PickleType

from .meta import SnorkelBase, snorkel_postgres


class Context(SnorkelBase):
    """
    A piece of content from which Candidates are composed.
    """
    __tablename__ = 'context'
    id            = Column(Integer, primary_key=True)
    type          = Column(String, nullable=False)
    stable_id     = Column(String, unique=True, nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': 'context',
        'polymorphic_on': type
    }

    def get_parent(self):
        raise NotImplementedError()

    def get_children(self):
        raise NotImplementedError()

    def get_sentence_generator(self):
        raise NotImplementedError()


class Document(Context):
    """
    A root Context.
    """
    __tablename__ = 'document'
    id            = Column(Integer, ForeignKey('context.id', ondelete='CASCADE'), primary_key=True)
    name          = Column(String, unique=True, nullable=False)
    text          = Column(String)
    meta          = Column(PickleType)

    __mapper_args__ = {
        'polymorphic_identity': 'document',
    }

    def get_parent(self):
        return None

    def get_children(self):
        return self.sentences

    def get_sentence_generator(self):
        for sentence in self.sentences:
            yield sentence

    def __repr__(self):
        return "Document " + str(self.name.encode('utf-8'))

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()


class Sentence(Context):
    """A sentence Context in a Document."""
    __tablename__ = 'sentence'
    id = Column(Integer, ForeignKey('context.id', ondelete='CASCADE'), primary_key=True)
    document_id = Column(Integer, ForeignKey('document.id', ondelete='CASCADE'))
    position = Column(Integer, nullable=False)
    document = relationship('Document', backref=backref('sentences', order_by=position, cascade='all, delete-orphan'), foreign_keys=document_id)
    text = Column(Text, nullable=False)
    if snorkel_postgres:
        words             = Column(postgresql.ARRAY(String), nullable=False)
        char_offsets      = Column(postgresql.ARRAY(Integer), nullable=False)
        abs_char_offsets  = Column(postgresql.ARRAY(Integer), nullable=False)
        lemmas            = Column(postgresql.ARRAY(String))
        pos_tags          = Column(postgresql.ARRAY(String))
        ner_tags          = Column(postgresql.ARRAY(String))
        dep_parents       = Column(postgresql.ARRAY(Integer))
        dep_labels        = Column(postgresql.ARRAY(String))
        entity_cids       = Column(postgresql.ARRAY(String))
        entity_types      = Column(postgresql.ARRAY(String))
    else:
        words             = Column(PickleType, nullable=False)
        char_offsets      = Column(PickleType, nullable=False)
        abs_char_offsets  = Column(PickleType, nullable=False)
        lemmas            = Column(PickleType)
        pos_tags          = Column(PickleType)
        ner_tags          = Column(PickleType)
        dep_parents       = Column(PickleType)
        dep_labels        = Column(PickleType)
        entity_cids       = Column(PickleType)
        entity_types      = Column(PickleType)

    __mapper_args__ = {
        'polymorphic_identity': 'sentence',
    }

    __table_args__ = (
        UniqueConstraint(document_id, position),
    )

    def get_parent(self):
        return self.document

    def get_children(self):
        return self.spans

    def _asdict(self):
        return {
            'id': self.id,
            'document': self.document,
            'position': self.position,
            'text': self.text,
            'words': self.words,
            'char_offsets': self.char_offsets,
            'lemmas': self.lemmas,
            'pos_tags': self.pos_tags,
            'ner_tags': self.ner_tags,
            'dep_parents': self.dep_parents,
            'dep_labels': self.dep_labels,
            'entity_cids': self.entity_cids,
            'entity_types': self.entity_types
        }

    def get_sentence_generator(self):
        yield self

    def __repr__(self):
        return "Sentence(%s,%s,%s)" % (self.document, self.position, self.text.encode('utf-8'))


class TemporaryContext(object):
    """
    A context which does not incur the overhead of a proper ORM-based Context object.
    The TemporaryContext class is specifically for the candidate extraction process, during which a CandidateSpace
    object will generate many TemporaryContexts, which will then be filtered by Matchers prior to materialization
    of Candidates and constituent Context objects.

    Every Context object has a corresponding TemporaryContext object from which it inherits.

    A TemporaryContext must have specified equality / set membership semantics, a stable_id for checking
    uniqueness against the database, and a promote() method which returns a corresponding Context object.
    """
    def __init__(self):
        self.id = None

    def load_id_or_insert(self, session):
        if self.id is None:
            stable_id = self.get_stable_id()
            id = session.execute(select([Context.id]).where(Context.stable_id == stable_id)).first()
            if id is None:
                self.id = session.execute(
                        Context.__table__.insert(),
                        {'type': self._get_table_name(), 'stable_id': stable_id}).inserted_primary_key[0]
                insert_args = self._get_insert_args()
                insert_args['id'] = self.id
                for (key, val) in insert_args.items():
                    if isinstance(val, list):
                        if snorkel_postgres:
                            insert_args[key] = val
                        else:
                            insert_args[key] = pickle.dumps(val) #NOTE: this works for sqlite, not Postgres
                session.execute(text(self._get_insert_query()), insert_args)
            else:
                self.id = id[0]

    def __eq__(self, other):
        raise NotImplementedError()

    def __ne__(self, other):
        raise NotImplementedError()

    def __hash__(self):
        raise NotImplementedError()

    def _get_polymorphic_identity(self):
        raise NotImplementedError()

    def get_stable_id(self):
        raise NotImplementedError()

    def _get_table_name(self):
        raise NotImplementedError()

    def _get_insert_query(self):
        raise NotImplementedError()

    def _get_insert_args(self):
        raise NotImplementedError()


class TemporarySpan(TemporaryContext):
    """The TemporaryContext version of Span"""
    def __init__(self, sentence, char_start, char_end, meta=None):
        super(TemporarySpan, self).__init__()
        self.sentence   = sentence  # The sentence Context of the Span
        self.char_end   = char_end
        self.char_start = char_start
        self.meta       = meta

    def __len__(self):
        return self.char_end - self.char_start + 1

    def __eq__(self, other):
        try:
            return self.sentence == other.sentence and self.char_start == other.char_start \
                and self.char_end == other.char_end
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return self.sentence != other.sentence or self.char_start != other.char_start \
                or self.char_end != other.char_end
        except AttributeError:
            return True

    def __hash__(self):
        return hash(self.sentence) + hash(self.char_start) + hash(self.char_end)

    def get_stable_id(self):
        return construct_stable_id(self.sentence, self._get_polymorphic_identity(), self.char_start, self.char_end)

    def _get_table_name(self):
        return 'span'

    def _get_polymorphic_identity(self):
        return 'span'

    def _get_insert_query(self):
        return """INSERT INTO span VALUES(:id, :sentence_id, :char_start, :char_end, :meta)"""

    def _get_insert_args(self):
        return {'sentence_id' : self.sentence.id,
                'char_start': self.char_start,
                'char_end'  : self.char_end,
                'meta'      : self.meta}

    def get_word_start(self):
        return self.char_to_word_index(self.char_start)

    def get_word_end(self):
        return self.char_to_word_index(self.char_end)

    def get_n(self):
        return self.get_word_end() - self.get_word_start() + 1

    def char_to_word_index(self, ci):
        """Given a character-level index (offset), return the index of the **word this char is in**"""
        i = None
        for i, co in enumerate(self.sentence.char_offsets):
            if ci == co:
                return i
            elif ci < co:
                return i-1
        return i

    def word_to_char_index(self, wi):
        """Given a word-level index, return the character-level index (offset) of the word's start"""
        return self.sentence.char_offsets[wi]

    def get_attrib_tokens(self, a='words'):
        """Get the tokens of sentence attribute _a_ over the range defined by word_offset, n"""
        return self.sentence.__getattribute__(a)[self.get_word_start():self.get_word_end() + 1]

    def get_attrib_span(self, a, sep=" "):
        """Get the span of sentence attribute _a_ over the range defined by word_offset, n"""
        # NOTE: Special behavior for words currently (due to correspondence with char_offsets)
        if a == 'words':
            return self.sentence.text[self.char_start:self.char_end + 1]
        else:
            return sep.join(self.get_attrib_tokens(a))

    def get_span(self, sep=" "):
        return self.get_attrib_span('words', sep)

    def __contains__(self, other_span):
        return (self.sentence == other_span.sentence
            and other_span.char_start >= self.char_start
            and other_span.char_end <= self.char_end)

    def __getitem__(self, key):
        """
        Slice operation returns a new candidate sliced according to **char index**
        Note that the slicing is w.r.t. the candidate range (not the abs. sentence char indexing)
        """
        if isinstance(key, slice):
            char_start = self.char_start if key.start is None else self.char_start + key.start
            if key.stop is None:
                char_end = self.char_end
            elif key.stop >= 0:
                char_end = self.char_start + key.stop - 1
            else:
                char_end = self.char_end + key.stop
            return self._get_instance(char_start=char_start, char_end=char_end, sentence=self.sentence)
        else:
            raise NotImplementedError()

    def __repr__(self):
        return '%s("%s", sentence=%s, chars=[%s,%s], words=[%s,%s])' \
            % (self.__class__.__name__, self.get_span().encode('utf-8'), self.sentence.id, self.char_start, self.char_end,
               self.get_word_start(), self.get_word_end())

    def _get_instance(self, **kwargs):
        return TemporarySpan(**kwargs)


class Span(Context, TemporarySpan):
    """
    A span of characters, identified by Context id and character-index start, end (inclusive).

    char_offsets are **relative to the Context start**
    """
    __tablename__ = 'span'
    id            = Column(Integer, ForeignKey('context.id', ondelete='CASCADE'), primary_key=True)
    sentence_id   = Column(Integer, ForeignKey('context.id', ondelete='CASCADE'))
    char_start    = Column(Integer, nullable=False)
    char_end      = Column(Integer, nullable=False)
    meta          = Column(PickleType)

    __table_args__ = (
        UniqueConstraint(sentence_id, char_start, char_end),
    )

    __mapper_args__ = {
        'polymorphic_identity': 'span',
        'inherit_condition': (id == Context.id)
    }

    sentence = relationship('Context', backref=backref('spans', cascade='all, delete-orphan'), foreign_keys=sentence_id)

    def get_parent(self):
        return self.sentence

    def get_children(self):
        return None

    def _get_instance(self, **kwargs):
        return Span(**kwargs)

    # We redefine these to use default semantics, overriding the operators inherited from TemporarySpan
    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return self is not other

    def __hash__(self):
        return id(self)


def split_stable_id(stable_id):
    """
    Split stable id, returning:
        * Document (root) stable ID
        * Context polymorphic type
        * Character offset start, end *relative to document start*
    Returns tuple of four values.
    """
    split1 = stable_id.split('::')
    if len(split1) == 2:
        split2 = split1[1].split(':')
        if len(split2) == 3:
            return split1[0], split2[0], int(split2[1]), int(split2[2])
    raise ValueError("Malformed stable_id:", stable_id)


def construct_stable_id(parent_context, polymorphic_type, relative_char_offset_start, relative_char_offset_end):
    """Contruct a stable ID for a Context given its parent and its character offsets relative to the parent"""
    doc_id, _, parent_doc_char_start, _ = split_stable_id(parent_context.stable_id)
    start = parent_doc_char_start + relative_char_offset_start
    end   = parent_doc_char_start + relative_char_offset_end
    return "%s::%s:%s:%s" % (doc_id, polymorphic_type, start, end)



from builtins import str
from builtins import range
from fonduer.lf_helpers import *
from fonduer.utils_table import min_row_diff, min_col_diff, num_rows, num_cols
from fonduer.config import settings

FEAT_PRE = 'TAB_'
DEF_VALUE = 1

unary_tablelib_feats = {}
binary_tablelib_feats = {}


def get_table_feats(candidates):
    candidates = candidates if isinstance(candidates, list) else [candidates]
    for candidate in candidates:
        args = candidate.get_contexts()
        if not (isinstance(args[0], TemporarySpan)):
            raise ValueError("Accepts Span-type arguments, %s-type found." % type(candidate))

        # Unary candidates
        if len(args) == 1:
            span = args[0]
            if span.stable_id not in unary_tablelib_feats:
                unary_tablelib_feats[span.stable_id] = set()
                for f, v in tablelib_unary_features(span):
                    unary_tablelib_feats[span.stable_id].add((f, v))

            for f, v in unary_tablelib_feats[span.stable_id]:
                yield candidate.id, FEAT_PRE + f, v

        # Binary candidates
        elif len(args) == 2:
            span1, span2 = args
            if span1.sentence.is_tabular() or span2.sentence.is_tabular():
                for span, pre in [(span1, "e1_"), (span2, "e2_")]:
                    if span.stable_id not in unary_tablelib_feats:
                        unary_tablelib_feats[span.stable_id] = set()
                        for f, v in tablelib_unary_features(span):
                            unary_tablelib_feats[span.stable_id].add((f, v))

                    for f, v in unary_tablelib_feats[span.stable_id]:
                        yield candidate.id, FEAT_PRE + pre + f, v

                if candidate.id not in binary_tablelib_feats:
                    binary_tablelib_feats[candidate.id] = set()
                    for f, v in tablelib_binary_features(span1, span2):
                        binary_tablelib_feats[candidate.id].add((f, v))

                for f, v in binary_tablelib_feats[candidate.id]:
                    yield candidate.id, FEAT_PRE + f, v
        else:
            raise NotImplementedError("Only handles unary and binary candidates currently")


def tablelib_unary_features(span):
    """
    Table-/structure-related features for a single span
    """
    if not span.sentence.is_tabular(): return
    phrase = span.sentence
    for attrib in settings.featurization.table.unary_features.attrib:
        for ngram in get_cell_ngrams(span,
                                     n_max=settings.featurization.table.unary_features.get_cell_ngrams.max,
                                     attrib=attrib):
            yield "CELL_%s_[%s]" % (attrib.upper(), ngram), DEF_VALUE
        for row_num in range(phrase.row_start, phrase.row_end + 1):
            yield "ROW_NUM_[%s]" % row_num, DEF_VALUE
        for col_num in range(phrase.col_start, phrase.col_end + 1):
            yield "COL_NUM_[%s]" % col_num, DEF_VALUE
        # NOTE: These two features could be accounted for by HTML_ATTR in structural features
        yield "ROW_SPAN_[%d]" % num_rows(phrase), DEF_VALUE
        yield "COL_SPAN_[%d]" % num_cols(phrase), DEF_VALUE
        for axis in ['row', 'col']:
            for ngram in get_head_ngrams(span, axis,
                                         n_max=settings.featurization.table.unary_features.get_head_ngrams.max,
                                         attrib=attrib):
                yield "%s_HEAD_%s_[%s]" % (axis.upper(), attrib.upper(), ngram), 1
        for ngram in get_row_ngrams(span, n_max=settings.featurization.table.unary_features.get_row_ngrams.max,
                                    attrib=attrib):
            yield "ROW_%s_[%s]" % (attrib.upper(), ngram), DEF_VALUE
        for ngram in get_col_ngrams(span, n_max=settings.featurization.table.unary_features.get_col_ngrams.max,
                                    attrib=attrib):
            yield "COL_%s_[%s]" % (attrib.upper(), ngram), DEF_VALUE
            # TODO:
            # for ngram in get_row_ngrams(span, n_max=2, attrib=attrib, direct=False, infer=True):
            #     yield "ROW_INFERRED_%s_[%s]" % (attrib.upper(), ngram), DEF_VALUE
            # for ngram in get_col_ngrams(span, n_max=2, attrib=attrib, direct=False, infer=True):
            #     yield "COL_INFERRED_%s_[%s]" % (attrib.upper(), ngram), DEF_VALUE


def tablelib_binary_features(span1, span2):
    """
    Table-/structure-related features for a pair of spans
    """
    if span1.sentence.is_tabular() and span2.sentence.is_tabular():
        if span1.sentence.table == span2.sentence.table:
            yield u"SAME_TABLE", DEF_VALUE
            if span1.sentence.cell is not None and span2.sentence.cell is not None:
                row_diff = min_row_diff(span1.sentence, span2.sentence,
                                        absolute=settings.featurization.table.binary_features.min_row_diff.absolute)
                col_diff = min_col_diff(span1.sentence, span2.sentence,
                                        absolute=settings.featurization.table.binary_features.min_col_diff.absolute)
                yield u"SAME_TABLE_ROW_DIFF_[%s]" % row_diff, DEF_VALUE
                yield u"SAME_TABLE_COL_DIFF_[%s]" % col_diff, DEF_VALUE
                yield u"SAME_TABLE_MANHATTAN_DIST_[%s]" % str(abs(row_diff) + abs(col_diff)), DEF_VALUE
                if span1.sentence.cell == span2.sentence.cell:
                    yield u"SAME_CELL", DEF_VALUE
                    yield u"WORD_DIFF_[%s]" % (span1.get_word_start() - span2.get_word_start()), DEF_VALUE
                    yield u"CHAR_DIFF_[%s]" % (span1.char_start - span2.char_start), DEF_VALUE
                    if span1.sentence == span2.sentence:
                        yield u"SAME_PHRASE", DEF_VALUE
        else:
            if span1.sentence.cell is not None and span2.sentence.cell is not None:
                yield u"DIFF_TABLE", DEF_VALUE
                row_diff = min_row_diff(span1.sentence, span2.sentence,
                                        absolute=settings.featurization.table.binary_features.min_row_diff.absolute)
                col_diff = min_col_diff(span1.sentence, span2.sentence,
                                        absolute=settings.featurization.table.binary_features.min_col_diff.absolute)
                yield u"DIFF_TABLE_ROW_DIFF_[%s]" % row_diff, DEF_VALUE
                yield u"DIFF_TABLE_COL_DIFF_[%s]" % col_diff, DEF_VALUE
                yield u"DIFF_TABLE_MANHATTAN_DIST_[%s]" % str(abs(row_diff) + abs(col_diff)), DEF_VALUE



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

import sys
import requests

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry


class Parser(object):

    def __init__(self, name, encoding='utf-8'):
        self.name = name
        self.encoding = encoding

    def to_unicode(self, text):
        '''
        Convert char encoding to unicode
        :param text:
        :return:
        '''
        if sys.version_info[0] < 3:
            text_alt = text.encode('utf-8', 'error')
            text_alt = text_alt.decode('string_escape', errors='ignore')
            text_alt = text_alt.decode('utf-8')
            return text_alt
        else:
            return text

    def connect(self):
        '''
        Return connection object for this parser type
        :return:
        '''
        raise NotImplemented

    def close(self):
        '''
        Kill this parser
        :return:
        '''
        raise NotImplemented


class ParserConnection(object):
    '''
    Default connection object assumes local parser object
    '''
    def __init__(self, parser):
        self.parser = parser

    def _connection(self):
        raise NotImplemented

    def parse(self, document, text):
        return self.parser.parse(document, text)


class URLParserConnection(ParserConnection):
    '''
    URL parser connection
    '''
    def __init__(self, parser, retries=5):
        self.retries = retries
        self.parser = parser
        self.request = self._connection()

    def _connection(self):
        '''
        Enables retries to cope with CoreNLP server boot-up latency.
        See: http://stackoverflow.com/a/35504626

        Create a new object per connection to make multiprocessing threadsafe.

        :return:
        '''
        requests_session = requests.Session()
        retries = Retry(total=self.retries,
                        connect=self.retries,
                        read=self.retries,
                        backoff_factor=0.1,
                        status_forcelist=[500, 502, 503, 504])

        # Mac OS bug -- without this setting multiprocessing requests will fail
        # when the server has boot-up latency associated with model loading
        # See: http://stackoverflow.com/questions/30453152/python-multiprocessing-and-requests
        if sys.platform in ['darwin']:
            requests_session.trust_env = False
        requests_session.mount('http://', HTTPAdapter(max_retries=retries))
        return requests_session

    def post(self, url, data, allow_redirects=True):
        '''

        :param url:
        :param data:
        :param allow_redirects:
        :param timeout:
        :return:
        '''
        resp = self.request.post(url, data=data, allow_redirects=allow_redirects)
        return resp.content.strip()

    def parse(self, document, text):
        '''
        Return parse generator
        :param document:
        :param text:
        :return:
        '''
        return self.parser.parse(document, text, self)








from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

from traitlets import Unicode, Int, List
import getpass
import ipywidgets as widgets
import os
import warnings

from fonduer.snorkel.models import GoldLabel, StableLabel, GoldLabelKey
try:
    from IPython.core.display import display, Javascript
except:
    raise Exception("This module must be run in IPython.")
warnings.filterwarnings("ignore", category=DeprecationWarning)

HOME = os.environ['FONDUERHOME']


# PAGE LAYOUT TEMPLATES
LI_HTML = u"""
<li class="list-group-item" data-toggle="tooltip" data-placement="top" title="{context_id}">{data}</li>
"""

PAGE_HTML = u"""
<div class="viewer-page" id="viewer-page-{pid}"{etc}>
    <ul class="list-group">{data}</ul>
</div>
"""


class Viewer(widgets.DOMWidget):
    """
    Generic object for viewing and labeling Candidate objects in their rendered Contexts.
    """
    _view_name         = Unicode('ViewerView').tag(sync=True)
    _view_module       = Unicode('viewer').tag(sync=True)
    cids               = List().tag(sync=True)
    html               = Unicode('<h3>Error!</h3>').tag(sync=True)
    _labels_serialized = Unicode().tag(sync=True)
    _selected_cid      = Int().tag(sync=True)

    def __init__(self, candidates, session, gold=[], n_per_page=3, height=225, annotator_name=None):
        """
        Initializes a Viewer.

        The Viewer uses the keyword argument annotator_name to define a AnnotatorLabelKey with that name.

        :param candidates: A Python container of Candidates (e.g., not a CandidateSet, but candidate_set.candidates)
        :param session: The SnorkelSession for the database backend
        :param gold: Optional, Python container of Candidates that are know to have positive labels
        :param n_per_page: Optional, number of Contexts to display per page
        :param height: Optional, the height in pixels of the Viewer
        :param annotator_name: Name of the human using the Viewer, for saving their work. Defaults to system username.
        """
        super(Viewer, self).__init__()
        self.session = session

        # By default, use the username as annotator name
        name = annotator_name if annotator_name is not None else getpass.getuser()

        # Sets up the AnnotationKey to use
        self.annotator = self.session.query(GoldLabelKey).filter(GoldLabelKey.name == name).first()
        if self.annotator is None:
            self.annotator = GoldLabelKey(name=name)
            session.add(self.annotator)
            session.commit()

        # Viewer display configs
        self.n_per_page = n_per_page
        self.height     = height

        # Note that the candidates are not necessarily commited to the DB, so they *may not have* non-null ids
        # Hence, we index by their position in this list
        # We get the sorted candidates and all contexts required, either from unary or binary candidates
        self.gold       = list(gold)
        self.candidates = sorted(list(candidates), key=lambda c : c[0].char_start)
        self.contexts   = list(set(c[0].get_parent() for c in self.candidates + self.gold))

        # If committed, sort contexts by id
        try:
            self.contexts = sorted(self.contexts, key=lambda c : c.id)
        except:
            pass

        # Loads existing annotations
        self.annotations        = [None] * len(self.candidates)
        self.annotations_stable = [None] * len(self.candidates)
        init_labels_serialized  = []
        for i, candidate in enumerate(self.candidates):

            # First look for the annotation in the primary annotations table
            existing_annotation = self.session.query(GoldLabel) \
                .filter(GoldLabel.key == self.annotator) \
                .filter(GoldLabel.candidate == candidate) \
                .first()
            if existing_annotation is not None:
                self.annotations[i] = existing_annotation
                if existing_annotation.value == 1:
                    value_string = 'true'
                elif existing_annotation.value == -1:
                    value_string = 'false'
                else:
                    raise ValueError(str(existing_annotation) +
                                     ' has value not in {1, -1}, which Viewer does not support.')
                init_labels_serialized.append(str(i) + '~~' + value_string)

                # If the annotator label is in the main table, also get its stable version
                context_stable_ids = '~~'.join([c.stable_id for c in candidate.get_contexts()])
                existing_annotation_stable = self.session.query(StableLabel) \
                                                 .filter(StableLabel.context_stable_ids == context_stable_ids)\
                                                 .filter(StableLabel.annotator_name == name).one_or_none()

                # If stable version is not available, create it here
                # NOTE: This is for versioning issues, should be removed?
                if existing_annotation_stable is None:
                    context_stable_ids         = '~~'.join([c.stable_id for c in candidate.get_contexts()])
                    existing_annotation_stable = StableLabel(context_stable_ids=context_stable_ids,\
                                                             annotator_name=self.annotator.name,\
                                                             split=candidate.split,\
                                                             value=existing_annotation.value)
                    self.session.add(existing_annotation_stable)
                    self.session.commit()

                self.annotations_stable[i] = existing_annotation_stable

        self._labels_serialized = ','.join(init_labels_serialized)

        # Configures message handler
        self.on_msg(self.handle_label_event)

        # display js, construct html and pass on to widget model
        self.render()

    def _tag_span(self, html, cids, gold=False):
        """
        Create the span around a segment of the context associated with one or more candidates / gold annotations
        """
        classes  = ['candidate'] if len(cids) > 0 else []
        classes += ['gold-annotation'] if gold else []
        classes += list(map(str, cids))

        # Scrub for non-ascii characters; replace with ?
        return u'<span class="{classes}">{html}</span>'.format(classes=' '.join(classes), html=html)

    def _tag_context(self, context, candidates, gold):
        """Given the raw context, tag the spans using the generic _tag_span method"""
        raise NotImplementedError()

    def render(self):
        """Renders viewer pane"""
        cids = []

        # Iterate over pages of contexts
        pid   = 0
        pages = []
        N     = len(self.contexts)
        for i in range(0, N, self.n_per_page):
            page_cids = []
            lis       = []
            for j in range(i, min(N, i + self.n_per_page)):
                context = self.contexts[j]

                # Get the candidates in this context
                candidates = [c for c in self.candidates if c[0].get_parent() == context]
                gold = [g for g in self.gold if g.context_id == context.id]

                # Construct the <li> and page view elements
                li_data = self._tag_context(context, candidates, gold)
                lis.append(LI_HTML.format(data=li_data, context_id=context.id))
                page_cids.append([self.candidates.index(c) for c in candidates])

            # Assemble the page...
            pages.append(PAGE_HTML.format(
                pid=pid,
                data=''.join(lis),
                etc=' style="display: block;"' if i == 0 else ''
            ))
            cids.append(page_cids)
            pid += 1

        # Render in primary Viewer template
        self.cids = cids
        self.html = open(HOME+'/viewer/viewer.html').read() % (self.height, ''.join(pages))
        display(Javascript(open(HOME + '/viewer/viewer.js').read()))

    def _get_labels(self):
        """
        De-serialize labels from Javascript widget, map to internal candidate id, and return as list of tuples
        """
        LABEL_MAP = {'true':1, 'false':-1}
        labels    = [x.split('~~') for x in self._labels_serialized.split(',') if len(x) > 0]
        vals      = [(int(cid), LABEL_MAP.get(l, 0)) for cid,l in labels]
        return vals

    def handle_label_event(self, _, content, buffers):
        """
        Handles label event by persisting new label
        """
        if content.get('event', '') == 'set_label':
            cid = content.get('cid', None)
            value = content.get('value', None)
            if value is True:
                value = 1
            elif value is False:
                value = -1
            else:
                raise ValueError('Unexpected label returned from widget: ' + str(value) +
                                 '. Expected values are True and False.')

            # If label already exists, just update value (in both AnnotatorLabel and StableLabel)
            if self.annotations[cid] is not None:
                if self.annotations[cid].value != value:
                    self.annotations[cid].value        = value
                    self.annotations_stable[cid].value = value
                    self.session.commit()

            # Otherwise, create a AnnotatorLabel *and a StableLabel*
            else:
                candidate = self.candidates[cid]

                # Create AnnotatorLabel
                self.annotations[cid] = GoldLabel(key=self.annotator, candidate=candidate, value=value)
                self.session.add(self.annotations[cid])

                # Create StableLabel
                context_stable_ids           = '~~'.join([c.stable_id for c in candidate.get_contexts()])
                self.annotations_stable[cid] = StableLabel(context_stable_ids=context_stable_ids,\
                                                           annotator_name=self.annotator.name,\
                                                           value=value,\
                                                           split=candidate.split)
                self.session.add(self.annotations_stable[cid])
                self.session.commit()

        elif content.get('event', '') == 'delete_label':
            cid = content.get('cid', None)
            self.session.delete(self.annotations[cid])
            self.annotations[cid] = None
            self.session.delete(self.annotations_stable[cid])
            self.annotations_stable[cid] = None
            self.session.commit()

    def get_selected(self):
        return self.candidates[self._selected_cid]


class SentenceNgramViewer(Viewer):
    """Viewer for Sentence objects and candidate Spans within them"""
    def __init__(self, candidates, session, gold=[], n_per_page=3, height=225, annotator_name=None):
        super(SentenceNgramViewer, self).__init__(candidates, session, gold=gold, n_per_page=n_per_page, height=height, annotator_name=annotator_name)

    def _is_subspan(self, s, e, c):
        return s >= c.char_start and e <= c.char_end

    def _tag_context(self, sentence, candidates, gold):
        """Tag **potentially overlapping** spans of text, at the character-level"""
        s = sentence.text

        # First, split the sentence into the *smallest* single-candidate chunks
        try:
            both = [c[0] for c in candidates] + [c[1] for c in candidates] \
                        + [g[0] for g in gold] + [g[1] for g in gold]
        except:
            both = [c[0] for c in candidates] + [g[0] for g in gold]
        splits = sorted(list(set([b.char_start for b in both] + [b.char_end + 1 for b in both] + [0, len(s)])))

        # For each chunk, add cid if subset of candidate span, tag if gold, and produce span
        html = ""
        for i in range(len(splits)-1):
            start  = splits[i]
            end    = splits[i+1] - 1

            # Handle both unary and binary candidates
            try:
                # For binary candidates, add classes for both the candidate ID and unary span identifiers
                cids0  = [self.candidates.index(c) for c in candidates if self._is_subspan(start, end, c[0])]
                cids0 += ['%s-0' % cid for cid in cids0]
                cids1  = [self.candidates.index(c) for c in candidates if self._is_subspan(start, end, c[1])]
                cids1 += ['%s-1' % cid for cid in cids1]
                cids   = cids0 + cids1

                # Handle gold...
                gcids = [self.gold.index(g) for g in gold if
                         self._is_subspan(start, end, g[0]) or self._is_subspan(start, end, g[1])]
            except:
                cids  = [self.candidates.index(c) for c in candidates if self._is_subspan(start, end, c[0])]
                gcids = [self.gold.index(g) for g in gold if self._is_subspan(start, end, g[0])]

            html += self._tag_span(s[start:end+1], cids, gold=len(gcids) > 0)
        return html



from builtins import map
from builtins import object
from sqlalchemy import Column, String, Integer, Text, ForeignKey, UniqueConstraint
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship, backref
from sqlalchemy.sql import text
from sqlalchemy.types import PickleType

from fonduer.snorkel.models.context import Context
from fonduer.snorkel.models.context import Document
from fonduer.snorkel.models.context import TemporaryContext
from fonduer.snorkel.models.context import TemporarySpan
from fonduer.snorkel.models.context import split_stable_id
from fonduer.snorkel.models.meta import snorkel_postgres

INT_ARRAY_TYPE = postgresql.ARRAY(Integer) if snorkel_postgres else PickleType
STR_ARRAY_TYPE = postgresql.ARRAY(String)  if snorkel_postgres else PickleType


class Webpage(Document):
    """
    Declares name for storage table.
    """
    __tablename__ = 'webpage'
    id            = Column(Integer, ForeignKey('document.id', ondelete='CASCADE'), primary_key=True)
    # Connects NewType records to generic Context records
    url           = Column(String)
    host          = Column(String)
    page_type     = Column(String)
    raw_content   = Column(String)
    crawltime     = Column(String)
    all           = Column(String)

    # Polymorphism information for SQLAlchemy
    __mapper_args__ = {
        'polymorphic_identity': 'webpage',
    }

    # Rest of class definition here
    def __repr__(self):
        return "Webpage(id: %s..., url: %s...)" % (self.name[:10].encode('utf-8'), self.url[8:23].encode('utf-8'))


class Table(Context):
    """A table Context in a Document."""
    __tablename__ = 'table'
    id            = Column(Integer, ForeignKey('context.id', ondelete='CASCADE'), primary_key=True)
    document_id   = Column(Integer, ForeignKey('document.id'))
    position      = Column(Integer, nullable=False)
    document      = relationship('Document', backref=backref('tables', order_by=position, cascade='all, delete-orphan'), foreign_keys=document_id)

    __mapper_args__ = {
        'polymorphic_identity': 'table',
    }

    __table_args__ = (
        UniqueConstraint(document_id, position),
    )

    def __repr__(self):
        return "Table(Doc: %s, Position: %s)" % (self.document.name.encode('utf-8'), self.position)

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()


class Figure(Context):
    """A figure Context in a Document."""
    __tablename__ = 'figure'
    id            = Column(Integer, ForeignKey('context.id', ondelete='CASCADE'), primary_key=True)
    document_id   = Column(Integer, ForeignKey('document.id', ondelete='CASCADE'))
    position      = Column(Integer, nullable=False)
    document      = relationship('Document', backref=backref('figures', order_by=position, cascade='all, delete-orphan'), foreign_keys=document_id)
    url = Column(String)

    __mapper_args__ = {
        'polymorphic_identity': 'figure',
    }

    __table_args__ = (
        UniqueConstraint(document_id, position),
    )

    def __repr__(self):
        return "Figure(Doc: %s, Position: %s, Url: %s)" % (self.document.name.encode('utf-8'), self.position, self.url)

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()


class Cell(Context):
    """A cell Context in a Document."""
    __tablename__ = 'cell'
    id            = Column(Integer, ForeignKey('context.id', ondelete='CASCADE'), primary_key=True)
    document_id   = Column(Integer, ForeignKey('document.id'))
    table_id      = Column(Integer, ForeignKey('table.id'))
    position      = Column(Integer, nullable=False)
    document      = relationship('Document', backref=backref('cells', order_by=position, cascade='all, delete-orphan'), foreign_keys=document_id)
    table         = relationship('Table', backref=backref('cells', order_by=position, cascade='all, delete-orphan'), foreign_keys=table_id)
    row_start = Column(Integer)
    row_end = Column(Integer)
    col_start = Column(Integer)
    col_end = Column(Integer)
    html_tag = Column(Text)
    if snorkel_postgres:
        html_attrs = Column(postgresql.ARRAY(String))
        # html_anc_tags = Column(postgresql.ARRAY(String))
        # html_anc_attrs = Column(postgresql.ARRAY(String))
    else:
        html_attrs = Column(PickleType)
        # html_anc_tags = Column(PickleType)
        # html_anc_attrs = Column(PickleType)

    __mapper_args__ = {
        'polymorphic_identity': 'cell',
    }

    __table_args__ = (
        UniqueConstraint(document_id, table_id, position),
    )

    def __repr__(self):
        return ("Cell(Doc: %s, Table: %s, Row: %s, Col: %s, Pos: %s)" %
                (self.document.name.encode('utf-8'),
                 self.table.position,
                 tuple(set([self.row_start, self.row_end])),
                 tuple(set([self.col_start, self.col_end])),
                 self.position))

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()


class PhraseMixin(object):
    """A phrase Context in a Document."""
    def is_lingual(self):
        return False

    def is_visual(self):
        return False

    def is_tabular(self):
        return False

    def is_structural(self):
        return False

    def __repr__(self):
        return ("Phrase (Doc: %s, Index: %s, Text: %s)" %
                (self.document.name.encode('utf-8'),
                 self.phrase_idx,
                 self.text.encode('utf-8')))


class LingualMixin(object):
    """A collection of lingual attributes."""
    lemmas      = Column(STR_ARRAY_TYPE)
    pos_tags    = Column(STR_ARRAY_TYPE)
    ner_tags    = Column(STR_ARRAY_TYPE)
    dep_parents = Column(INT_ARRAY_TYPE)
    dep_labels  = Column(STR_ARRAY_TYPE)

    def is_lingual(self):
        return self.lemmas is not None

    def __repr__(self):
        return ("LingualPhrase (Doc: %s, Index: %s, Text: %s)" %
                (self.document.name.encode('utf-8'),
                 self.phrase_idx,
                 self.text.encode('utf-8')))


class TabularMixin(object):
    """A collection of tabular attributes."""
    @declared_attr
    def table_id(cls):
        return Column('table_id', ForeignKey('table.id'))

    @declared_attr
    def table(cls):
        return relationship('Table', backref=backref('phrases', cascade='all, delete-orphan'), foreign_keys=lambda: cls.table_id)

    @declared_attr
    def cell_id(cls):
        return Column('cell_id', ForeignKey('cell.id'))

    @declared_attr
    def cell(cls):
        return relationship('Cell', backref=backref('phrases', cascade='all, delete-orphan'), foreign_keys=lambda: cls.cell_id)

    row_start = Column(Integer)
    row_end   = Column(Integer)
    col_start = Column(Integer)
    col_end   = Column(Integer)
    position  = Column(Integer)

    def is_tabular(self):
        return self.table is not None

    def is_cellular(self):
        return self.cell is not None

    def __repr__(self):
        rows = tuple([self.row_start, self.row_end]) if self.row_start != self.row_end else self.row_start
        cols = tuple([self.col_start, self.col_end]) if self.col_start != self.col_end else self.col_start
        return (
            "TabularPhrase (Doc: %s, Table: %s, Row: %s, Col: %s, Index: %s, Text: %s)" %
            (self.document.name.encode('utf-8'),
            (lambda: cls.table).position,
            rows,
            cols,
            self.phrase_idx,
            self.text.encode('utf-8'))
        )


class VisualMixin(object):
    """A collection of visual attributes."""
    page   = Column(INT_ARRAY_TYPE)
    top    = Column(INT_ARRAY_TYPE)
    left   = Column(INT_ARRAY_TYPE)
    bottom = Column(INT_ARRAY_TYPE)
    right  = Column(INT_ARRAY_TYPE)

    def is_visual(self):
        return self.page is not None and self.page[0] is not None

    def __repr__(self):
        return ("VisualPhrase (Doc: %s, Page: %s, (T,B,L,R): (%d,%d,%d,%d), Text: %s)" %
            (self.document.name.encode('utf-8'),
            self.page, self.top, self.bottom, self.left, self.right,
            self.text.encode('utf-8')))


class StructuralMixin(object):
    """A collection of structural attributes."""
    xpath      = Column(String)
    html_tag   = Column(String)
    html_attrs = Column(STR_ARRAY_TYPE)

    def is_structural(self):
        return self.html_tag is not None

    def __repr__(self):
        return ("StructuralPhrase (Doc: %s, Tag: %s, Text: %s)" %
            (self.document.name.encode('utf-8'),
            self.html_tag,
            self.text.encode('utf-8')))


# PhraseMixin must come last in arguments to not ovewrite is_* methods
# class Phrase(Context, StructuralMixin, PhraseMixin): # Memex variant
class Phrase(Context, TabularMixin, LingualMixin, VisualMixin, StructuralMixin, PhraseMixin):
    """A Phrase subclass with Lingual, Tabular, Visual, and HTML attributes."""
    __tablename__    = 'phrase'
    id               = Column(Integer, ForeignKey('context.id', ondelete='CASCADE'), primary_key=True)
    document_id      = Column(Integer, ForeignKey('document.id'))
    document         = relationship('Document', backref=backref('phrases', cascade='all, delete-orphan'), foreign_keys=document_id)
    phrase_num       = Column(Integer, nullable=False)  # unique Phrase number per document
    text             = Column(Text, nullable=False)
    words            = Column(STR_ARRAY_TYPE)
    char_offsets     = Column(INT_ARRAY_TYPE)
    entity_cids      = Column(STR_ARRAY_TYPE)
    entity_types     = Column(STR_ARRAY_TYPE)
    abs_char_offsets = Column(INT_ARRAY_TYPE)

    __mapper_args__ = {
        'polymorphic_identity': 'phrase',
    }

    __table_args__ = (
        UniqueConstraint(document_id, phrase_num),
    )

    def __repr__(self):
        if self.is_tabular():
            rows = tuple([self.row_start, self.row_end]) \
                   if self.row_start != self.row_end else self.row_start
            cols = tuple([self.col_start, self.col_end]) \
                   if self.col_start != self.col_end else self.col_start
            return ("Phrase (Doc: %s, Table: %s, Row: %s, Col: %s, Index: %s, Text: %s)" %
                   (self.document.name.encode('utf-8'),
                    self.table.position,
                    rows,
                    cols,
                    self.position,
                    self.text.encode('utf-8')))
        else:
            return ("Phrase (Doc: %s, Index: %s, Text: %s)" %
                (self.document.name.encode('utf-8'),
                self.phrase_num,
                self.text.encode('utf-8')))

    def _asdict(self):
        return {
            # base
            'id': self.id,
            # 'document': self.document,
            'phrase_num': self.phrase_num,
            'text': self.text,
            'entity_cids': self.entity_cids,
            'entity_types': self.entity_types,
            # tabular
            # 'table': self.table,
            # 'cell': self.cell,
            'row_start': self.row_start,
            'row_end': self.row_end,
            'col_start': self.col_start,
            'col_end': self.col_end,
            'position': self.position,
            # lingual
            'words': self.words,
            'char_offsets': self.char_offsets,
            'lemmas': self.lemmas,
            'pos_tags': self.pos_tags,
            'ner_tags': self.ner_tags,
            'dep_parents': self.dep_parents,
            'dep_labels': self.dep_labels,
            # visual
            'page': self.page,
            'top': self.top,
            'bottom': self.bottom,
            'left': self.left,
            'right': self.right
        }

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()


class TemporaryImplicitSpan(TemporarySpan):
    """The TemporaryContext version of ImplicitSpan"""
    def __init__(self, sentence, char_start, char_end, expander_key, position,
            text, words, lemmas, pos_tags, ner_tags, dep_parents, dep_labels,
            page, top, left, bottom, right, meta=None):
        super(TemporarySpan, self).__init__()
        self.sentence       = sentence  # The sentence Context of the Span
        self.char_start     = char_start
        self.char_end       = char_end
        self.expander_key   = expander_key
        self.position       = position
        self.text           = text
        self.words          = words
        self.lemmas         = lemmas
        self.pos_tags       = pos_tags
        self.ner_tags       = ner_tags
        self.dep_parents    = dep_parents
        self.dep_labels     = dep_labels
        self.page           = page
        self.top            = top
        self.left           = left
        self.bottom         = bottom
        self.right          = right
        self.meta           = meta

    def __len__(self):
        return sum(map(len, self.words))

    def __eq__(self, other):
        try:
            return (self.sentence == other.sentence and
                    self.char_start == other.char_start and
                    self.char_end == other.char_end and
                    self.expander_key == other.expander_key and
                    self.position == other.position)
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return (self.sentence != other.sentence or
                    self.char_start != other.char_start or
                    self.char_end != other.char_end or
                    self.expander_key != other.expander_key or
                    self.position != other.position)
        except AttributeError:
            return True

    def __hash__(self):
        return (hash(self.sentence) + hash(self.char_start) + hash(self.char_end)
                + hash(self.expander_key) + hash(self.position))

    def get_stable_id(self):
        doc_id, _, parent_doc_char_start, _ = split_stable_id(self.sentence.stable_id)
        # return (construct_stable_id(self.sentence, self._get_polymorphic_identity(), self.char_start, self.char_end)
        #     + ':%s:%s' % (self.expander_key, self.position))
        return '%s::%s:%s:%s:%s:%s' % (
            self.sentence.document.name,
            self._get_polymorphic_identity(),
            parent_doc_char_start + self.char_start,
            parent_doc_char_start + self.char_end,
            self.expander_key,
            self.position)

    def _get_table_name(self):
        return 'implicit_span'

    def _get_polymorphic_identity(self):
        return 'implicit_span'

    def _get_insert_query(self):
        return """INSERT INTO implicit_span VALUES(
            :id,
            :sentence_id,
            :char_start,
            :char_end,
            :expander_key,
            :position,
            :text,
            :words,
            :lemmas,
            :pos_tags,
            :ner_tags,
            :dep_parents,
            :dep_labels,
            :page,
            :top,
            :left,
            :bottom,
            :right,
            :meta)"""

    def _get_insert_args(self):
        return {'sentence_id'   : self.sentence.id,
                'char_start'    : self.char_start,
                'char_end'      : self.char_end,
                'expander_key'  : self.expander_key,
                'position'      : self.position,
                'text'          : self.text,
                'words'         : self.words,
                'lemmas'        : self.lemmas,
                'pos_tags'      : self.pos_tags,
                'ner_tags'      : self.ner_tags,
                'dep_parents'   : self.dep_parents,
                'dep_labels'    : self.dep_labels,
                'page'          : self.page,
                'top'           : self.top,
                'left'          : self.left,
                'bottom'        : self.bottom,
                'right'         : self.right,
                'meta'          : self.meta
                }

    def get_attrib_tokens(self, a='words'):
        """Get the tokens of sentence attribute _a_ over the range defined by word_offset, n"""
        return self.__getattribute__(a)

    def get_attrib_span(self, a, sep=" "):
        """Get the span of sentence attribute _a_ over the range defined by word_offset, n"""
        if a == 'words':
            return self.text
        else:
            return sep.join(self.get_attrib_tokens(a))

    def __getitem__(self, key):
        """Slice operation returns a new candidate sliced according to **char index**.

        Note that the slicing is w.r.t. the candidate range (not the abs. sentence char indexing)
        """
        if isinstance(key, slice):
            char_start = self.char_start if key.start is None else self.char_start + key.start
            if key.stop is None:
                char_end = self.char_end
            elif key.stop >= 0:
                char_end = self.char_start + key.stop - 1
            else:
                char_end = self.char_end + key.stop
            return self._get_instance(sentence=self.sentence, char_start=char_start, char_end=char_end, expander_key=self.expander_key,
                position=self.position, text=text, words=self.words, lemmas=self.lemmas, pos_tags=self.pos_tags,
                ner_tags=self.ner_tags, dep_parents=self.dep_parents, dep_labels=self.dep_labels,
                page=self.page, top=self.top, left=self.left, bottom=self.bottom, right=self.right, meta=self.meta)
        else:
            raise NotImplementedError()

    def __repr__(self):
        return '%s("%s", sentence=%s, words=[%s,%s], position=[%s])' \
            % (self.__class__.__name__, self.get_span().encode('utf-8'), self.sentence.id,
               self.get_word_start(), self.get_word_end(), self.position)

    def _get_instance(self, **kwargs):
        return TemporaryImplicitSpan(**kwargs)


class ImplicitSpan(Context, TemporaryImplicitSpan):
    """A span of characters that may not have appeared verbatim in the source text.

    It is identified by Context id, character-index start and end (inclusive),
    as well as a key representing what 'expander' function drew the ImplicitSpan
    from an  existing Span, and a position (where position=0 corresponds to the
    first ImplicitSpan produced from the expander function).

    The character-index start and end point to the segment of text that was
    expanded to produce the ImplicitSpan.
    """
    __tablename__ = 'implicit_span'
    id            = Column(Integer, ForeignKey('context.id', ondelete='CASCADE'), primary_key=True)
    sentence_id   = Column(Integer, ForeignKey('context.id', ondelete='CASCADE'), primary_key=True)
    char_start    = Column(Integer, nullable=False)
    char_end      = Column(Integer, nullable=False)
    expander_key  = Column(String, nullable=False)
    position      = Column(Integer, nullable=False)
    text          = Column(String)
    if snorkel_postgres:
        words       = Column(postgresql.ARRAY(String), nullable=False)
        lemmas      = Column(postgresql.ARRAY(String))
        pos_tags    = Column(postgresql.ARRAY(String))
        ner_tags    = Column(postgresql.ARRAY(String))
        dep_parents = Column(postgresql.ARRAY(Integer))
        dep_labels  = Column(postgresql.ARRAY(String))
        page        = Column(postgresql.ARRAY(Integer))
        top         = Column(postgresql.ARRAY(Integer))
        left        = Column(postgresql.ARRAY(Integer))
        bottom      = Column(postgresql.ARRAY(Integer))
        right       = Column(postgresql.ARRAY(Integer))
    else:
        words       = Column(PickleType, nullable=False)
        lemmas      = Column(PickleType)
        pos_tags    = Column(PickleType)
        ner_tags    = Column(PickleType)
        dep_parents = Column(PickleType)
        dep_labels  = Column(PickleType)
        page        = Column(PickleType)
        top         = Column(PickleType)
        left        = Column(PickleType)
        bottom      = Column(PickleType)
        right       = Column(PickleType)
    meta = Column(PickleType)

    __table_args__ = (
        UniqueConstraint(sentence_id, char_start, char_end, expander_key, position),
    )

    __mapper_args__ = {
        'polymorphic_identity': 'implicit_span',
        'inherit_condition': (id == Context.id)
    }

    sentence = relationship('Context', backref=backref('implicit_spans', cascade='all, delete-orphan'), foreign_keys=sentence_id)

    def _get_instance(self, **kwargs):
        return ImplicitSpan(**kwargs)

    # We redefine these to use default semantics, overriding the operators inherited from TemporarySpan
    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return self is not other

    def __hash__(self):
        return id(self)


class TemporaryImage(TemporaryContext):
    """The TemporaryContext version of Figure"""
    def __init__(self, figure):
        super(TemporaryImage, self).__init__()
        self.figure   = figure  # The figure Context

    def __len__(self):
        return 1

    def __eq__(self, other):
        try:
            return self.figure.position == other.figure.position \
                and self.figure.url == other.figure.url
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return self.figure.position != other.figure.position \
                or self.figure.url != other.figure.url
        except AttributeError:
            return True

    def __contains__(self, other_span):
        return self.__eq__(other_span)

    def __hash__(self):
        return hash(self.figure)

    def get_stable_id(self):
        return "%s::%s:%s" % (self.figure.document.id, self._get_polymorphic_identity(), self.figure.position)

    def _get_table_name(self):
        return 'image'

    def _get_polymorphic_identity(self):
        return 'image'

    def _get_insert_query(self):
        return """INSERT INTO image VALUES(:id, :document_id, :image_id, :url)"""

    def _get_insert_args(self):
        return {'document_id': self.figure.document.id,
                'image_id' : self.figure.position,
                'url'      : self.figure.url}

    def __repr__(self):
        return '%s(document=%s, position=%s, url=%s)' \
            % (self.__class__.__name__, self.figure.document.name.encode('utf-8'), \
                self.figure.position, self.figure.url)

    def _get_instance(self, **kwargs):
        return TemporaryImage(**kwargs)


class Image(Context, TemporaryImage):
    """
    A candidate of figure, identified by Context id and position.
    """
    __tablename__ = 'image'
    id            = Column(Integer, ForeignKey('context.id', ondelete='CASCADE'), primary_key=True)
    document_id   = Column(Integer, ForeignKey('document.id', ondelete='CASCADE'))
    position      = Column(Integer, nullable=False)
    url           = Column(String)
    document      = relationship('Document', backref=backref('images', order_by=position, cascade='all, delete-orphan'), foreign_keys=document_id)


    __table_args__ = (
        UniqueConstraint(document_id, position),
    )

    __mapper_args__ = {
        'polymorphic_identity': 'image',
    }

    def __repr__(self):
        return "Image(Doc: %s, Position: %s, Url: %s)" % (self.document.name.encode('utf-8'), self.position, self.url)

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()



"""
Subpackage for Snorkel machine learning modules.
"""
from __future__ import absolute_import

from .utils import *
from .disc_models.rnn import reRNN, TagRNN, TextRNN
from .disc_models.logistic_regression import (
	LogisticRegression, SparseLogisticRegression
)
from .gen_learning import (
    GenerativeModel,
    GenerativeModelWeights
)


from builtins import map
from builtins import range
from copy import deepcopy
from itertools import product
from sqlalchemy.sql import select

from fonduer.models import TemporaryImage
from fonduer.snorkel.candidates import CandidateSpace, Ngrams
from fonduer.snorkel.models import Candidate
from fonduer.snorkel.models.context import Document
from fonduer.snorkel.udf import UDF, UDFRunner

class CandidateExtractor(UDFRunner):
    """An operator to extract Candidate objects from a Context.

    :param candidate_class: The type of relation to extract, defined using
                            :func:`snorkel.models.candidate_subclass <snorkel.models.candidate.candidate_subclass>`
    :param cspaces: one or list of :class:`CandidateSpace` objects, one for each relation argument. Defines space of
                    Contexts to consider
    :param matchers: one or list of :class:`snorkel.matchers.Matcher` objects, one for each relation argument. Only tuples of
                     Contexts for which each element is accepted by the corresponding Matcher will be returned as Candidates
    :param candidate_filter: an optional function for filtering out candidates which returns a Boolean expressing whether or not
                      the candidate should be instantiated.
    :param self_relations: Boolean indicating whether to extract Candidates that relate the same context.
                           Only applies to binary relations. Default is False.
    :param nested_relations: Boolean indicating whether to extract Candidates that relate one Context with another
                             that contains it. Only applies to binary relations. Default is False.
    :param symmetric_relations: Boolean indicating whether to extract symmetric Candidates, i.e., rel(A,B) and rel(B,A),
                                where A and B are Contexts. Only applies to binary relations. Default is True.
    """

    def __init__(self, candidate_class, cspaces, matchers, candidate_filter=None, self_relations=False, nested_relations=False, symmetric_relations=True):
        """Initialize the CandidateExtractor."""
        super(CandidateExtractor, self).__init__(CandidateExtractorUDF,
                                                 candidate_class=candidate_class,
                                                 cspaces=cspaces,
                                                 matchers=matchers,
                                                 candidate_filter=candidate_filter,
                                                 self_relations=self_relations,
                                                 nested_relations=nested_relations,
                                                 symmetric_relations=symmetric_relations)

    def apply(self, xs, split=0, **kwargs):
        """Call the CandidateExtractorUDF."""
        super(CandidateExtractor, self).apply(xs, split=split, **kwargs)

    def clear(self, session, split, **kwargs):
        """Delete Candidates from given split the database."""
        session.query(Candidate).filter(Candidate.split == split).delete()


class CandidateExtractorUDF(UDF):
    """UDF for performing candidate extraction."""

    def __init__(self, candidate_class, cspaces, matchers, candidate_filter, self_relations, nested_relations, symmetric_relations, **kwargs):
        """Initialize the CandidateExtractorUDF."""
        self.candidate_class     = candidate_class
        self.candidate_spaces    = cspaces if type(cspaces) in [list, tuple] else [cspaces]
        self.matchers            = matchers if type(matchers) in [list, tuple] else [matchers]
        self.candidate_filter    = candidate_filter
        self.nested_relations    = nested_relations
        self.self_relations      = self_relations
        self.symmetric_relations = symmetric_relations

        # Check that arity is same
        if len(self.candidate_spaces) != len(self.matchers):
            raise ValueError("Mismatched arity of candidate space and matcher.")
        else:
            self.arity = len(self.candidate_spaces)

        # Make sure the candidate spaces are different so generators aren't expended!
        self.candidate_spaces = list(map(deepcopy, self.candidate_spaces))

        # Preallocates internal data structures
        self.child_context_sets = [None] * self.arity
        for i in range(self.arity):
            self.child_context_sets[i] = set()

        super(CandidateExtractorUDF, self).__init__(**kwargs)

    def apply(self, context, clear, split, **kwargs):
        """Extract candidates from the given Context.

        Here, we define a context as a Phrase.
        :param context:
        :param clear:
        :param split: Which split to use.
        """
        # Generate TemporaryContexts that are children of the context using the candidate_space and filtered
        # by the Matcher
        for i in range(self.arity):
            self.child_context_sets[i].clear()
            for tc in self.matchers[i].apply(self.candidate_spaces[i].apply(self.session, context)):
                tc.load_id_or_insert(self.session)
                self.child_context_sets[i].add(tc)

        # Generates and persists candidates
        candidate_args = {'split': split}
        for args in product(*[enumerate(child_contexts) for child_contexts in self.child_context_sets]):

            # Apply candidate_filter if one was given
            # Accepts a tuple of Context objects (e.g., (Span, Span))
            # (candidate_filter returns whether or not proposed candidate passes throttling condition)
            if self.candidate_filter:
                if not self.candidate_filter(tuple(args[i][1] for i in range(self.arity))):
                    continue

            # TODO: Make this work for higher-order relations
            if self.arity == 2:
                ai, a = args[0]
                bi, b = args[1]

                # Check for self-joins, "nested" joins (joins from span to its subspan), and flipped duplicate
                # "symmetric" relations
                if not self.self_relations and a == b:
                    continue
                elif not self.nested_relations and (a in b or b in a):
                    continue
                elif not self.symmetric_relations and ai > bi:
                    continue

            # Assemble candidate arguments
            for i, arg_name in enumerate(self.candidate_class.__argnames__):
                candidate_args[arg_name + '_id'] = args[i][1].id

            # Checking for existence
            if not clear:
                q = select([self.candidate_class.id])
                for key, value in list(candidate_args.items()):
                    q = q.where(getattr(self.candidate_class, key) == value)
                candidate_id = self.session.execute(q).first()
                if candidate_id is not None:
                    continue

            # Add Candidate to session
            yield self.candidate_class(**candidate_args)


class OmniNgrams(Ngrams):
    """
    Defines the space of candidates.

    Defines the space of candidates as all n-grams (n <= n_max) in a Document _x_,
    divided into Phrases inside of html elements (such as table cells).
    """

    def __init__(self, n_max=5, split_tokens=['-', '/']):
        """
        Initialize OmniNgrams.
        """
        Ngrams.__init__(self, n_max=n_max, split_tokens=split_tokens)

    def apply(self, session, context):
        """
        Generate OmniNgrams from a Document by parsing all of its Phrases.
        """
        if not isinstance(context, Document):
            raise TypeError("Input Contexts to OmniNgrams.apply() must be of type Document")

        doc = session.query(Document).filter(Document.id == context.id).one()
        for phrase in doc.phrases:
            for ts in Ngrams.apply(self, phrase):
                yield ts


class OmniFigures(CandidateSpace):
    """
    Defines the space of candidates as all figures in a Document _x_,
    indexing by **position offset**.
    """

    def __init__(self, type=None):
        """
        Initialize OmniFigures.

        Only support figure type filter.
        """
        CandidateSpace.__init__(self)
        if type is not None:
            self.type=type.strip().lower()
        self.type = None

    def apply(self, session, context):
        """
        Generate OmniFigures from a Document by parsing all of its Figures.
        """
        if not isinstance(context, Document):
            raise TypeError("Input Contexts to OmniFigures.apply() must be of type Document")

        doc = session.query(Document).filter(Document.id == context.id).one()
        for figure in doc.figures:
            if self.type is None or figure.url.lower().endswith(self.type):
                yield TemporaryImage(figure)



from __future__ import print_function
from __future__ import division
from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import str
from builtins import range
from builtins import object
import http.client
import os
import re
import subprocess
import warnings
from collections import OrderedDict, defaultdict

import numpy as np
import pandas as pd
from bs4 import BeautifulSoup
from editdistance import eval as editdist  # Alternative library: python-levenshtein


class VisualLinker(object):
    def __init__(self, time=False, verbose=False):
        self.pdf_file = None
        self.verbose = verbose
        self.time = time
        self.coordinate_map = None
        self.pdf_word_list = None
        self.html_word_list = None
        self.links = None
        self.pdf_dim = None
        delimiters = u"([\(\)\,\?\u2212\u201C\u201D\u2018\u2019\u00B0\*\']|(?<!http):|\.$|\.\.\.)"
        self.separators = re.compile(delimiters)

    def parse_visual(self, document_name, phrases, pdf_path):
        self.phrases = phrases
        self.pdf_file = pdf_path + document_name + '.pdf'
        if not os.path.isfile(self.pdf_file):
            self.pdf_file = self.pdf_file[:-3] + "PDF"
        try:
            self.extract_pdf_words()
        except RuntimeError as e:
            warnings.warn(e.message, RuntimeWarning)
            return
        self.extract_html_words()
        self.link_lists(search_max=200)
        for phrase in self.update_coordinates():
            yield phrase

    def extract_pdf_words(self):
        num_pages = subprocess.check_output(
            "pdfinfo '{}' | grep -a Pages | sed 's/[^0-9]*//'".format(
                self.pdf_file),
            shell=True)
        pdf_word_list = []
        coordinate_map = {}
        for i in range(1, int(num_pages) + 1):
            html_content = subprocess.check_output(
                "pdftotext -f {} -l {} -bbox-layout '{}' -".format(
                    str(i), str(i), self.pdf_file),
                shell=True)
            soup = BeautifulSoup(html_content, "html.parser")
            pages = soup.find_all('page')
            pdf_word_list_i, coordinate_map_i = self._coordinates_from_HTML(
                pages[0], i)
            pdf_word_list += pdf_word_list_i
            # update coordinate map
            coordinate_map.update(coordinate_map_i)
        self.pdf_word_list = pdf_word_list
        self.coordinate_map = coordinate_map
        if len(self.pdf_word_list) == 0:
            raise RuntimeError(
                "Words could not be extracted from PDF: %s" % self.pdf_file)
        # take last page dimensions
        page_width, page_height = int(float(pages[0].get('width'))), int(
            float(pages[0].get('height')))
        self.pdf_dim = (page_width, page_height)
        if self.verbose:
            print("Extracted %d pdf words" % len(self.pdf_word_list))

    def _coordinates_from_HTML(self, page, page_num):
        pdf_word_list = []
        coordinate_map = {}
        block_coordinates = {}
        blocks = page.find_all('block')
        i = 0  # counter for word_id in page_num
        for block in blocks:
            x_min_block = int(float(block.get('xmin')))
            y_min_block = int(float(block.get('ymin')))
            lines = block.find_all('line')
            for line in lines:
                y_min_line = int(float(line.get('ymin')))
                y_max_line = int(float(line.get('ymax')))
                words = line.find_all("word")
                for word in words:
                    xmin = int(float(word.get('xmin')))
                    xmax = int(float(word.get('xmax')))
                    for content in self.separators.split(word.getText()):
                        if len(content) > 0:  # Ignore empty characters
                            word_id = (page_num, i)
                            pdf_word_list.append((word_id, content))
                            coordinate_map[word_id] = (page_num, y_min_line,
                                                       xmin, y_max_line, xmax)
                            block_coordinates[word_id] = (y_min_block,
                                                          x_min_block)
                            i += 1
        # sort pdf_word_list by page, block top then block left, top, then left
        pdf_word_list = sorted(
            pdf_word_list,
            key=
            lambda word_id__: block_coordinates[word_id__[0]] + coordinate_map[word_id__[0]][1:3]
        )
        return pdf_word_list, coordinate_map

    def extract_html_words(self):
        html_word_list = []
        for phrase in self.phrases:
            for i, word in enumerate(phrase.words):
                html_word_list.append(((phrase.stable_id, i), word))
        self.html_word_list = html_word_list
        if self.verbose:
            print("Extracted %d html words" % len(self.html_word_list))

    def link_lists(self, search_max=100, edit_cost=20, offset_cost=1):
        # NOTE: there are probably some inefficiencies here from rehashing words
        # multiple times, but we're not going to worry about that for now

        def link_exact(l, u):
            l, u, L, U = get_anchors(l, u)
            html_dict = defaultdict(list)
            pdf_dict = defaultdict(list)
            for i, (_, word) in enumerate(self.html_word_list[l:u]):
                if html_to_pdf[l + i] is None:
                    html_dict[word].append(l + i)
            for j, (_, word) in enumerate(self.pdf_word_list[L:U]):
                if pdf_to_html[L + j] is None:
                    pdf_dict[word].append(L + j)
            for word, html_list in list(html_dict.items()):
                pdf_list = pdf_dict[word]
                if len(html_list) == len(pdf_list):
                    for k in range(len(html_list)):
                        html_to_pdf[html_list[k]] = pdf_list[k]
                        pdf_to_html[pdf_list[k]] = html_list[k]

        def link_fuzzy(i):
            (_, word) = self.html_word_list[i]
            l = u = i
            l, u, L, U = get_anchors(l, u)
            offset = int(L + float(i - l) / (u - l) * (U - L))
            searchIndices = np.clip(offset + search_order, 0, M - 1)
            cost = [0] * search_max
            for j, k in enumerate(searchIndices):
                other = self.pdf_word_list[k][1]
                if (word.startswith(other) or word.endswith(other)
                        or other.startswith(word) or other.endswith(word)):
                    html_to_pdf[i] = k
                    return
                else:
                    cost[j] = int(editdist(
                        word, other)) * edit_cost + j * offset_cost
            html_to_pdf[i] = searchIndices[np.argmin(cost)]
            return

        def get_anchors(l, u):
            while l >= 0 and html_to_pdf[l] is None:
                l -= 1
            while u < N and html_to_pdf[u] is None:
                u += 1
            if l < 0:
                l = 0
                L = 0
            else:
                L = html_to_pdf[l]
            if u >= N:
                u = N
                U = M
            else:
                U = html_to_pdf[u]
            return l, u, L, U

        def display_match_counts():
            matches = sum([
                html_to_pdf[i] is not None and self.html_word_list[i][1] ==
                self.pdf_word_list[html_to_pdf[i]][1]
                for i in range(len(self.html_word_list))
            ])
            total = len(self.html_word_list)
            print("({:d}/{:d}) = {:.2f}".format(matches, total,
                                                matches / total))
            return matches

        N = len(self.html_word_list)
        M = len(self.pdf_word_list)
        assert (N > 0 and M > 0)
        html_to_pdf = [None] * N
        pdf_to_html = [None] * M
        search_radius = search_max // 2

        # first pass: global search for exact matches
        link_exact(0, N)
        if self.verbose:
            print("Global exact matching:")
            display_match_counts()

        # second pass: local search for exact matches
        for i in range(((N + 2) // search_radius) + 1):
            link_exact(
                max(0, i * search_radius - search_radius),
                min(N, i * search_radius + search_radius))
        if self.verbose:
            print("Local exact matching:")
            display_match_counts()

        # third pass: local search for approximate matches
        search_order = np.array(
            [(-1)**(i % 2) * (i // 2) for i in range(1, search_max + 1)])
        for i in range(len(html_to_pdf)):
            if html_to_pdf[i] is None:
                link_fuzzy(i)
        if self.verbose:
            print("Local approximate matching:")
            display_match_counts()

        # convert list to dict
        matches = sum([
            html_to_pdf[i] is not None and
            self.html_word_list[i][1] == self.pdf_word_list[html_to_pdf[i]][1]
            for i in range(len(self.html_word_list))
        ])
        total = len(self.html_word_list)
        if self.verbose:
            print("Linked {:d}/{:d} ({:.2f}) html words exactly".format(
                matches, total, matches / total))
        self.links = OrderedDict((self.html_word_list[i][0],
                                  self.pdf_word_list[html_to_pdf[i]][0])
                                 for i in range(len(self.html_word_list)))

    def link_lists_old(self,
                       search_max=200,
                       editCost=20,
                       offsetCost=1,
                       offsetInertia=5):
        DEBUG = False
        if DEBUG:
            offsetHist = []
            jHist = []
            editDistHist = 0
        offset = self._calculate_offset(self.html_word_list,
                                        self.pdf_word_list,
                                        max((search_max // 10), 5), search_max)
        offsets = [offset] * offsetInertia
        searchOrder = np.array(
            [(-1)**(i % 2) * (i // 2) for i in range(1, search_max + 1)])
        links = OrderedDict()
        for i, a in enumerate(self.html_word_list):
            j = 0
            searchIndices = np.clip(offset + searchOrder, 0,
                                    len(self.pdf_word_list) - 1)
            jMax = len(searchIndices)
            matched = False
            # Search first for exact matches
            while not matched and j < jMax:
                b = self.pdf_word_list[searchIndices[j]]
                if a[1] == b[1]:
                    links[a[0]] = b[0]
                    matched = True
                    offsets[i % offsetInertia] = searchIndices[j] + 1
                    offset = int(np.median(offsets))
                    if DEBUG:
                        jHist.append(j)
                        offsetHist.append(offset)
                j += 1
            # If necessary, search for min edit distance
            if not matched:
                cost = [0] * search_max
                for k, m in enumerate(searchIndices):
                    cost[k] = (
                        editdist(a[1], self.pdf_word_list[m][1]) * editCost +
                        k * offsetCost)
                nearest = np.argmin(cost)
                links[a[0]] = self.pdf_word_list[searchIndices[nearest]][0]
                if DEBUG:
                    jHist.append(nearest)
                    offsetHist.append(searchIndices[nearest])
                    editDistHist += 1
        if DEBUG:
            print(offsetHist)
            print(jHist)
            print(editDistHist)
            self.offsetHist = offsetHist
        self.links = links
        if self.verbose:
            print("Linked {:d} words to {:d} bounding boxes".format(
                len(self.html_word_list), len(self.pdf_word_list)))

    def _calculate_offset(self, listA, listB, seedSize, maxOffset):
        wordsA = zip(*listA[:seedSize])[1]
        wordsB = zip(*listB[:maxOffset])[1]
        offsets = []
        for i in range(seedSize):
            try:
                offsets.append(wordsB.index(wordsA[i]) - i)
            except:
                pass
        return int(np.median(offsets))

    def display_links(self, max_rows=100):
        html = []
        pdf = []
        j = []
        for i, l in enumerate(self.links):
            html.append(self.html_word_list[i][1])
            for k, b in enumerate(self.pdf_word_list):
                if b[0] == self.links[self.html_word_list[i][0]]:
                    pdf.append(b[1])
                    j.append(k)
                    break
        assert (len(pdf) == len(html))

        total = 0
        match = 0
        for i, word in enumerate(html):
            total += 1
            if word == pdf[i]:
                match += 1
        print((match, total, match / total))

        data = {
            # 'i': range(len(self.links)),
            'html': html,
            'pdf': pdf,
            'j': j,
        }
        pd.set_option('display.max_rows', max_rows)
        print(pd.DataFrame(data, columns=['html', 'pdf', 'j']))
        pd.reset_option('display.max_rows')

    def update_coordinates(self):
        for phrase in self.phrases:
            (page, top, left, bottom, right) = list(
                zip(*[
                    self.coordinate_map[self.links[((phrase.stable_id), i)]]
                    for i in range(len(phrase.words))
                ]))
            phrase.page = list(page)
            phrase.top = list(top)
            phrase.left = list(left)
            phrase.bottom = list(bottom)
            phrase.right = list(right)
            yield phrase
        if self.verbose:
            print("Updated coordinates in database")
